"""
 Re-implementation of MUS-enumeration using MARCO.
"""

import cpmpy as cp
from cpmpy.transformations.get_variables import get_variables

from .utils import make_assump_model



def marco(soft, hard=[], solver="ortools", map_solver="ortools", return_mus=True, return_mcs=True, do_solution_hint=True):
    """
        Enumerating minimal unsatisfiable subsets (MUSes) and minimal correction sets (MCSes)
        of unsatisfiable constraints.
        
        Iteratively generates a subset of constraints (the seed) and checks whether it is SAT.
        When the seed is SAT, it is grown to an MSS to derive an MCS.
        This MCS is then added to the Map solver as a set to hit
        In case the seed is UNSAT, the seed is shrunk to a real MUS and returned.
        The Map-solver is instructed to not generated any superset of this MUS as next seeds

        Based on:
            Liffiton, Mark H., et al. "Fast, flexible MUS enumeration." Constraints 21 (2016): 223-250.
    
        :param: solver: name of a solver, must support assumptions (e.g, "ortools", "exact", "z3" or "pysat")
        :param: map_solver: the hitting-set (MAP) solver to use, ideally incremental such as "gurobi", "pysat" or "exact"
        :param: return_mus: whether the algorithm should return MUSes
        :param: return_mcs: whether the algorithm should return MCSes
        :param: do_solution_hint: when true, will favor large seeds generated by the map-solver, and hence more likely
                                     to return MUSes. Especially useful when `return_mus=True`.

    """

    assert hasattr(cp.SolverLookup.get(solver), "get_core"), "MARCO requires a solver that supports assumption variables"

    model, soft, assump = make_assump_model(soft, hard)
    dmap = dict(zip(assump, soft))
    s = cp.SolverLookup.get(solver, model)

    # map solver for computing hitting sets
    map_solver = cp.SolverLookup.get(map_solver)
    do_solution_hint = do_solution_hint and hasattr(map_solver, 'solution_hint')  # solver may not support solution hinting...

    map_solver += cp.any(assump)
    if do_solution_hint:
        hint = [1]*len(assump)
        map_solver.solution_hint(assump, hint) # we want large subsets, more likely to be a MUS

    deletion_order = {a : -len(get_variables(dmap[a])) for a in assump} # avoid recomputing

    while map_solver.solve():

        seed = [a for a in assump if a.value()]

        if s.solve(assumptions=seed) is True:
            # SAT, grow, to full MSS
            # Assumptions encode indicator constraints a -> c, find all true assumptions
            #    and those that could just as well be made true given the current solution
            mss = [a for a,c in zip(assump, soft) if a.value() or c.value()]
            for to_add in set(assump) - set(mss):
                if s.solve(assumptions=mss + [to_add]) is True:
                    mss.append(to_add)
            mcs = [a for a in assump if a not in frozenset(mss)] # take complement
            map_solver += cp.any(mcs) # block in map solver

            if return_mcs:
                yield "MCS", [dmap[a] for a in mcs]


        else: # UNSAT, shrink to MUS, re-use MUSX
            core = set(s.get_core())
            for c in sorted(core, key=deletion_order.get):
                if c not in core: # already removed
                    continue
                core.remove(c)
                if s.solve(assumptions=list(core)):
                    core.add(c)
                else: # UNSAT, shrink to new solver core (clause set refinement)
                    core = set(s.get_core())

            map_solver += ~cp.all(core) # block in map solver

            if return_mus:
                yield "MUS", [dmap[a] for a in core]


        # ensure solution hint is still active
        if do_solution_hint:
            map_solver.solution_hint(assump, hint)





"""
Benchmark solvers on XCSP3 Instances by replicating the XCSP3 competition.

A command-line tool for benchmarking constraint solvers on XCSP3 competition instances.
Supports parallel execution, time/memory limits, and solver configuration.

Required Arguments
------------------
--year : int
    The competition year (e.g., 2023).

--track : str
    The competition track (e.g., "CSP", "COP", "MiniCOP").

--solver : str
    The name of the solver to benchmark (e.g., "ortools", "exact", "choco").

Optional Arguments
------------------
--workers : int, default=4
    Number of parallel workers to use.

--time-limit : int, default=300
    Time limit in seconds per instance.

--mem-limit : int, default=8192
    Memory limit in megabytes per instance.

--output-dir : str, default='results'
    Directory where result CSV files will be saved.

--verbose
    If set, display full xcsp3 output during execution.

--intermediate
    If set, report intermediate solutions (if supported by the solver).
"""

import csv
import os
import signal
import subprocess
import time
import lzma
import sys
import argparse
import warnings
import traceback
import multiprocessing
from tqdm import tqdm
from pathlib import Path
from typing import Optional, Tuple
from io import StringIO
from datetime import datetime
from filelock import FileLock
from concurrent.futures import ThreadPoolExecutor

from cpmpy.tools.xcsp3.dataset import XCSP3Dataset
from cpmpy.tools.xcsp3.xcsp3_cpmpy import xcsp3_cpmpy, init_signal_handlers, ExitStatus

class Tee:
    """
    A stream-like object that duplicates writes to multiple underlying streams.
    """
    def __init__(self, *streams):
        """
        Arguments:
            *streams: Any number of file-like objects that implement a write() method,
                      such as sys.stdout, sys.stderr, or StringIO.
        """
        self.streams = streams

    def write(self, data):
        """
        Write data to all underlying streams.

        Args:
            data (str): The string to write.
        """
        for s in self.streams:
            s.write(data)

    def flush(self):
        """
        Flush all underlying streams to ensure all data is written out.
        """
        for s in self.streams:
            s.flush()

class PipeWriter:
    """
    Stdout wrapper for a multiprocessing pipe.
    """
    def __init__(self, conn):
        self.conn = conn
    def write(self, data):
        if data:  # avoid empty writes
            try:
                self.conn.send(data)
            except:
                pass
    def flush(self):
        pass  # no buffering


def xcsp3_wrapper(conn, kwargs, verbose):
    """
    Wraps a call to xcsp3_cpmpy as to correctly 
    forward stdout to the multiprocessing pipe (conn).
    Also sends a last status report though the pipe.

    Status report can be missing when process has been terminated by a SIGTERM.
    """
    
    original_stdout = sys.stdout

    pipe_writer = PipeWriter(conn)

    if not verbose:
        warnings.filterwarnings("ignore")
        sys.stdout = pipe_writer # only forward to pipe
    else:
        sys.stdout = Tee(original_stdout, pipe_writer) # forward to pipe and console

    try:
        init_signal_handlers() # configure OS signal handlers
        xcsp3_cpmpy(**kwargs)
        conn.send({"status": "ok"})
    except Exception as e: # capture exceptions and report in state
        tb_str = traceback.format_exc()
        conn.send({"status": "error", "exception": e, "traceback": tb_str})
    finally:
        sys.stdout = original_stdout
        conn.close()

# exec_args = (filename, metadata, solver, time_limit, mem_limit, output_file, verbose) 
def execute_instance(args: Tuple[str, dict, str, int, int, int, str, bool, bool, str]) -> None:
    """
    Solve a single XCSP3 instance and write results to file immediately.
    
    Args is a list of:
        filename: Path to the XCSP3 instance file
        metadata: Dictionary containing instance metadata (year, track, name)
        solver: Name of the solver to use
        time_limit: Time limit in seconds
        mem_limit: Memory limit in MB
        output_file: Path to the output CSV file
        verbose: Whether to show solver output
    """
    
    filename, metadata, solver, time_limit, mem_limit, cores, output_file, verbose, intermediate, checker_path = args

    # Fieldnames for the CSV file
    fieldnames = ['year', 'track', 'instance', 'solver',
                  'time_total', 'time_parse', 'time_model', 'time_post', 'time_solve',
                  'status', 'objective_value', 'solution', 'intermediate', 'checker_result']
    result = dict.fromkeys(fieldnames)  # init all fields to None
    result['year'] = metadata['year']
    result['track'] = metadata['track']
    result['instance'] = metadata['name'] 
    result['solver'] = solver

    # Decompress before timers start
    file_path = filename
    if str(filename).endswith(".lzma"):
        # Decompress the XZ file
        with lzma.open(filename, 'rt', encoding='utf-8') as f:
            xml_file = StringIO(f.read()) # read to memory-mapped file
            filename = xml_file
            
    # Start total timing
    total_start = time.time()
    
    # Call xcsp3 in separate process
    ctx = multiprocessing.get_context("spawn")
    parent_conn, child_conn = multiprocessing.Pipe() # communication pipe between processes
    process = ctx.Process(target=xcsp3_wrapper, args=(
                                                    child_conn, 
                                                      {
                                                          "benchname": filename, 
                                                          "solver": solver, 
                                                          "time_limit": time_limit, 
                                                          "mem_limit": mem_limit, 
                                                          "intermediate": intermediate, 
                                                          "force_mem_limit": True,
                                                          "time_buffer": 1,
                                                          "cores": cores,
                                                        }, 
                                                    verbose))
    process.start()
    process.join(timeout=time_limit)

    # Replicate competition convention on how jobs get terminated
    if process.is_alive():
        # Send sigterm to let process know it reached its time limit
        os.kill(process.pid, signal.SIGTERM)
        # 1 second grace period
        process.join(timeout=1)
        # Kill if still alive
        if process.is_alive():
            os.kill(process.pid, signal.SIGKILL)
            process.join()

    result['time_total'] = time.time() - total_start
       
    sol_time = None # For annotation intermediate solutions (when they were received)
    
    # Default status if nothing returned by subprocess
    # -> process exited prematurely due to sigterm
    status = {"status": "error", "exception": "sigterm"}

    # Parse the output to get status, solution and timings
    complete_solution = None
    while parent_conn.poll(timeout=1):
        line = parent_conn.recv()

        # Received a print statement from the subprocess
        if isinstance(line, str):
            if line.startswith('s '):
                result['status'] = line[2:].strip()
            elif line.startswith('v ') and result['solution'] is None:
                # only record first line, contains 'type' and 'cost'
                solution = line.split("\n")[0][2:].strip()
                result['solution'] = str(solution)
                complete_solution = line
                if "cost" in solution:
                    result['objective_value'] = solution.split('cost="')[-1][:-2]
            elif line.startswith('o '):
                obj = int(line[2:].strip())
                if result['intermediate'] is None:
                    result['intermediate'] = []
                result['intermediate'] += [(sol_time, obj)]
                result['objective_value'] = obj
                obj = None
            elif line.startswith('c Solution'):
                parts = line.split(', time = ')
                # Get solution time from comment for intermediate solution -> used for annotating 'o ...' lines
                sol_time = float(parts[-1].replace('s', '').rstrip())
            elif line.startswith('c took '):
                # Parse timing information
                parts = line.split(' seconds to ')
                if len(parts) == 2:
                    time_val = float(parts[0].replace('c took ', ''))
                    action = parts[1].strip()
                    if action.startswith('parse'):
                        result['time_parse'] = time_val
                    elif action.startswith('convert'):
                        result['time_model'] = time_val
                    elif action.startswith('post'):
                        result['time_post'] = time_val
                    elif action.startswith('solve'):
                        result['time_solve'] = time_val
        
        # Received a new status from the subprocess
        elif isinstance(line, dict):
            status = line

        else:
            raise()

    # Parse the exit status
    if status["status"] == "error":
        # Ignore timeouts
        if "TimeoutError" in repr(status["exception"]):
            pass
        # All other exceptions, put in solution field
        elif result['solution'] is None:
            result['status'] = ExitStatus.unknown.value
            result["solution"] = status["exception"]    

    if checker_path is not None and complete_solution is not None:
        checker_output, checker_time = run_solution_checker(
            JAR=checker_path,
            instance_location=file_path,
            out_file="'" + complete_solution.replace("\n\r", " ").replace("\n", " ").replace("v   ", "").replace("v ", "")+ "'",
            verbose=verbose,
            cpm_time=result.get('time_solve', 0)  # or total solve time you have
        )

        if checker_output is not None:
            result['checker_result'] = checker_output
        else:
            result['checker_result'] = None

    # Use a lock file to prevent concurrent writes
    lock_file = f"{output_file}.lock"
    lock = FileLock(lock_file)
    try:
        with lock:
            # Pre-check if file exists to determine if we need to write header
            write_header = not os.path.exists(output_file)

            with open(output_file, 'a', newline='') as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                if write_header:
                    writer.writeheader()
                writer.writerow(result)
    finally:
        # Optional: cleanup if the lock file somehow persists
        if os.path.exists(lock_file):
            try:
                os.remove(lock_file)
            except Exception:
                pass  # avoid crashing on cleanup


def run_solution_checker(JAR, instance_location, out_file, verbose, cpm_time):

    start = time.time()
    command = ["java", "-jar", JAR, "'" + str(instance_location) + "'" + " " + str(out_file)]
    command = " ".join(command)
    test_res_str = subprocess.run(command, capture_output=True, text=True, shell=True)
    checker_time = time.time() - start

    if verbose:
        for line in test_res_str.stdout.split("\n"):
            print("c " + line)
        print(f"c cpmpy time: {cpm_time}")
        print(f"c validation time: {checker_time}")
        print(f"c elapsed time: {cpm_time + checker_time}")
    
    return test_res_str.stdout.split("\n")[-2], checker_time


def xcsp3_benchmark(year: int, track: str, solver: str, workers: int = 1, 
                   time_limit: int = 300, mem_limit: Optional[int] = 4096, cores: int=1,
                   output_dir: str = 'results',
                   verbose: bool = False, intermediate: bool = False,
                   checker_path: Optional[str] = None) -> str:
    """
    Benchmark a solver on XCSP3 instances.
    
    Args:
        year (int): Competition year (e.g., 2023)
        track (str): Track type (e.g., COP, CSP, MiniCOP)
        solver (str): Solver name (e.g., ortools, exact, choco, ...)
        workers (int): Number of parallel workers
        time_limit (int): Time limit in seconds per instance
        mem_limit (int): Memory limit in MB per instance
        output_dir (str): Output directory for CSV files
        verbose (bool): Whether to show solver output
        
    Returns:
        str: Path to the output CSV file
    """
    # Create output directory
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    
    # Get current timestamp in a filename-safe format
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    
    # Define output file path with timestamp
    output_file = str(output_dir / f"xcsp3_{year}_{track}_{solver}_{timestamp}.csv")
    
    # Initialize dataset
    dataset = XCSP3Dataset(year=year, track=track, download=True)

    # Process instances in parallel
    with ThreadPoolExecutor(max_workers=workers) as executor:
        # Submit all tasks and track their futures
        futures = [executor.submit(execute_instance,  # below: args
                                   (filename, metadata, solver, time_limit, mem_limit, cores, output_file, verbose, intermediate, checker_path))
                   for filename, metadata in dataset]
        # Process results as they complete
        for i,future in enumerate(tqdm(futures, total=len(futures), desc=f"Running {solver}")):
            try:
                _ = future.result(timeout=time_limit+60)  # for cleanliness sake, result is empty
            except TimeoutError:
                pass
            except Exception as e:
                print(f"Job {i}: {dataset[i][1]['name']}, ProcessPoolExecutor caught: {e}")

        raise()
    
    return output_file

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Benchmark solvers on XCSP3 instances')
    parser.add_argument('--year', type=int, required=True, help='Competition year (e.g., 2023)')
    parser.add_argument('--track', type=str, required=True, help='Track type (e.g., COP, CSP, MiniCOP)')
    parser.add_argument('--solver', type=str, required=True, help='Solver name (e.g., ortools, exact, choco, ...)')
    parser.add_argument('--workers', type=int, default=4, help='Number of parallel workers')
    parser.add_argument('--time-limit', type=int, default=300, help='Time limit in seconds per instance')
    parser.add_argument('--mem-limit', type=int, default=8192, help='Memory limit in MB per instance')
    parser.add_argument('--cores', type=int, default=1, help='Number of cores to assign tp a single instance')
    parser.add_argument('--output-dir', type=str, default='results', help='Output directory for CSV files')
    parser.add_argument('--verbose', action='store_true', help='Show solver output')
    parser.add_argument('--intermediate', action='store_true', help='Report on intermediate solutions')
    parser.add_argument('--checker-path', type=str, default=None,
                    help='Path to the XCSP3 solution checker JAR file')
    
    
    args = parser.parse_args()

    if not args.verbose:
        warnings.filterwarnings("ignore")
    
    output_file = xcsp3_benchmark(**vars(args))
    print(f"Results added to {output_file}")



"""
Transforms flat constraints into linear constraints.

Linearized constraints have one of the following forms:

Linear comparison:
------------------
- ``LinExpr == Constant``
- ``LinExpr >= Constant``
- ``LinExpr <= Constant``

LinExpr can be any of:

- `NumVar`
- `sum`
- `wsum`

Indicator constraints:
----------------------

+------------------------------------+
| ``BoolVar -> LinExpr == Constant`` |
+------------------------------------+
| ``BoolVar -> LinExpr >= Constant`` |
+------------------------------------+
| ``BoolVar -> LinExpr <= Constant`` |
+------------------------------------+

========================================   ==============================================
``BoolVar -> GenExpr``                     (GenExpr.name in supported, GenExpr.is_bool()) 
``BoolVar -> GenExpr >= Var/Constant``     (GenExpr.name in supported, GenExpr.is_num())  
``BoolVar -> GenExpr <= Var/Constant``     (GenExpr.name in supported, GenExpr.is_num())  
``BoolVar -> GenExpr == Var/Constant``     (GenExpr.name in supported, GenExpr.is_num())  
========================================   ==============================================

Where ``BoolVar`` is a boolean variable or its negation.

General comparisons or expressions
-----------------------------------

============================  ==============================================
``GenExpr``                   (GenExpr.name in supported, GenExpr.is_bool())  
``GenExpr == Var/Constant``   (GenExpr.name in supported, GenExpr.is_num())  
``GenExpr <= Var/Constant``   (GenExpr.name in supported, GenExpr.is_num())  
``GenExpr >= Var/Constant``   (GenExpr.name in supported, GenExpr.is_num()) 
============================  ============================================== 



"""
import copy
import numpy as np
import cpmpy as cp
from cpmpy.transformations.get_variables import get_variables

from cpmpy.transformations.reification import only_implies


from .decompose_global import decompose_in_tree

from .flatten_model import flatten_constraint, get_or_make_var
from .normalize import toplevel_list, simplify_boolean
from .. import Abs
from ..exceptions import TransformationNotImplementedError

from ..expressions.core import Comparison, Expression, Operator, BoolVal
from ..expressions.globalconstraints import GlobalConstraint, DirectConstraint
from ..expressions.globalfunctions import GlobalFunction
from ..expressions.utils import is_bool, is_num, eval_comparison, get_bounds, is_true_cst, is_false_cst, is_int

from ..expressions.variables import _BoolVarImpl, boolvar, NegBoolView, _NumVarImpl, intvar

def linearize_constraint(lst_of_expr, supported={"sum","wsum"}, reified=False, csemap=None):
    """
    Transforms all constraints to a linear form.
    This function assumes all constraints are in 'flat normal form' with only boolean variables on the lhs of an implication.
    Only apply after :func:'cpmpy.transformations.flatten_model.flatten_constraint()' and :func:'cpmpy.transformations.reification.only_implies()'.

    Arguments:
        supported: which constraint and variable types are supported, i.e. `sum`, `and`, `or`, `alldifferent`
            :class:`~cpmpy.expressions.globalconstraints.AllDifferent` has a special linearization and is decomposed as such if not in `supported`.
            Any other unsupported global constraint should be decomposed using :func:`cpmpy.transformations.decompose_global.decompose_in_tree()`
        reified: whether the constraint is fully reified
    """

    newlist = []
    for cpm_expr in lst_of_expr:
        # Boolean literals are handled as trivial linears or unit clauses depending on `supported`
        if isinstance(cpm_expr, _BoolVarImpl):
            if "or" in supported:
                # post clause explicitly (don't use cp.any, which will just return the BoolVar)
                newlist.append(Operator("or", [cpm_expr]))
            elif isinstance(cpm_expr, NegBoolView):
                # might as well remove the negation
                newlist.append(sum([~cpm_expr]) <= 0)
            else: # positive literal
                newlist.append(sum([cpm_expr]) >= 1)

        # Boolean operators
        elif isinstance(cpm_expr, Operator) and cpm_expr.is_bool():
            # conjunction
            if cpm_expr.name == "and" and cpm_expr.name not in supported:
                newlist.append(sum(cpm_expr.args) >= len(cpm_expr.args))

            # disjunction
            elif cpm_expr.name == "or" and cpm_expr.name not in supported:
                newlist.append(sum(cpm_expr.args) >= 1)

            # reification
            elif cpm_expr.name == "->":
                # determine direction of implication
                cond, sub_expr = cpm_expr.args
                assert isinstance(cond, _BoolVarImpl), f"Linearization of {cpm_expr} is not supported, lhs of " \
                                                       f"implication must be boolvar. Apply `only_implies` before " \
                                                       f"calling `linearize_constraint`"

                if isinstance(cond, _BoolVarImpl) and isinstance(sub_expr, _BoolVarImpl):
                    # shortcut for BV -> BV, convert to disjunction and apply linearize on it
                    newlist.append(1 * cond + -1 * sub_expr <= 0)

                # BV -> LinExpr
                elif isinstance(cond, _BoolVarImpl):
                    lin_sub = linearize_constraint([sub_expr], supported=supported, reified=True, csemap=csemap)
                    # BV -> (C1 and ... and Cn) == (BV -> C1) and ... and (BV -> Cn)
                    indicator_constraints=[]
                    for lin in lin_sub:
                        if is_true_cst(lin):
                            continue
                        elif is_false_cst(lin):
                            indicator_constraints=[] # do not add any constraints
                            newlist+=linearize_constraint([~cond], supported=supported, csemap=csemap) # post linear version of unary constraint
                            break # do not need to add other
                        else:
                            indicator_constraints.append(cond.implies(lin)) # Add indicator constraint
                    newlist+=indicator_constraints

                    # ensure no new solutions are created
                    new_vars = set(get_variables(lin_sub)) - set(get_variables(sub_expr))
                    newlist += linearize_constraint([(~cond).implies(nv == nv.lb) for nv in new_vars], supported=supported, reified=reified, csemap=csemap)

            else: # supported operator
                newlist.append(cpm_expr)


        # comparisons
        elif isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args

            if lhs.name == "sub":
                # convert to wsum
                lhs = Operator("wsum", [[1, -1], [lhs.args[0], lhs.args[1]]])
                cpm_expr = eval_comparison(cpm_expr.name, lhs, rhs)

            if lhs.name == "-":
                lhs = Operator("wsum", [[-1], [lhs.args[0]]])
                cpm_expr = eval_comparison(cpm_expr.name, lhs, rhs)

            # linearize unsupported operators
            elif isinstance(lhs, Operator) and lhs.name not in supported:

                if lhs.name == "mul" and is_num(lhs.args[0]):
                    lhs = Operator("wsum",[[lhs.args[0]], [lhs.args[1]]])
                    cpm_expr = eval_comparison(cpm_expr.name, lhs, rhs)

                elif lhs.name == "pow" and "pow" not in supported:
                    if "mul" not in supported:
                        raise NotImplementedError("Cannot linearize power without multiplication")
                    if not is_int(lhs.args[1]) or lhs.args[1] < 0:
                        raise NotImplementedError("Cannot linearize power with non-integer or negative exponent")
                    # only `POW(b,n) == IV` supported, with n being a non-negative integer, post as b*b*...*b (n times) == IV
                    x, n = lhs.args
                    new_lhs = 1
                    for exp in range(n):
                        new_lhs, new_cons = get_or_make_var(x * new_lhs, csemap=csemap)
                        newlist.extend(new_cons)
                    cpm_expr = eval_comparison(cpm_expr.name, new_lhs, rhs)


                elif lhs.name == "mod" and "mod" not in supported:
                    if "mul" not in supported and not is_num(lhs.args[1]):
                        raise NotImplementedError("Cannot linearize modulo without multiplication")


                    if cpm_expr.name != "==":
                        new_rhs, newcons = get_or_make_var(lhs, csemap=csemap)
                        newlist.append(eval_comparison(cpm_expr.name, new_rhs, rhs))
                        newlist += linearize_constraint(newcons, supported=supported, reified=reified, csemap=csemap)
                        continue
                    else:
                        # mod != remainder after division because defined on integer div (rounding towards 0)
                        #   e.g., 7 % -5 = 2 and -7 % 5 = -2
                        # implement x % y == z as k * y + z == x with |z| < |y| and sign(x) = sign(z)
                        # https://marcelkliemannel.com/articles/2021/dont-confuse-integer-division-with-floor-division/
                        x, y = lhs.args
                        lby, uby = get_bounds(y)
                        if lby <= 0 <= uby:
                            raise ValueError("Attempting linearization of unsafe modulo, safen expression first (cpmpy/transformations/safen.py)")

                        # k * y + z == x
                        k = intvar(*get_bounds((x - rhs) // y))
                        mult_res, side_cons = get_or_make_var(k * y, csemap=csemap)
                        cpm_expr = (mult_res + rhs) == x
                        # |z| < |y|
                        abs_of_z, new_cons = get_or_make_var(abs(rhs), csemap=csemap)
                        side_cons += new_cons
                        # TODO: do the following in constructor of abs instead?
                        # we know y is strictly positive or negative due to safening.
                        if lby >= 0:
                            side_cons.append(abs_of_z < y)
                        if uby <= 0:
                            side_cons.append(abs_of_z < -y)
                        # sign(x) = sign(z)
                        lbx, ubx = get_bounds(x)
                        if lbx >= 0:
                            side_cons.append(rhs >= 0)
                        elif ubx <= 0:
                            side_cons.append(rhs <= 0)
                        else: # x can be pos or neg
                            x_is_pos = cp.boolvar()
                            x_is_neg = ~x_is_pos
                            side_cons += [
                                x_is_pos.implies(x >= 0), x_is_neg.implies(x < 0),
                                x_is_pos.implies(rhs >= 0), x_is_neg.implies(rhs <= 0)
                            ]

                        side_cons = toplevel_list(side_cons) # get rid of bools that may result from the above
                        newlist += linearize_constraint(side_cons, supported, reified=reified, csemap=csemap)

                elif lhs.name == 'div' and 'div' not in supported:
                    if "mul" not in supported:
                        raise NotImplementedError("Cannot linearize division without multiplication")

                    if cpm_expr.name != "==":
                        new_rhs, newcons = get_or_make_var(lhs, csemap=csemap)
                        newlist.append(eval_comparison(cpm_expr.name, new_rhs, rhs))
                        newlist += linearize_constraint(newcons, supported=supported, reified=reified, csemap=csemap)
                        continue

                    else:
                        # integer division, rounding towards zero
                        # x / y = z implemented as x = y * z + r with r the remainder and |r| < |y|
                        #      r can be positive or negative, so also ensure that |y| * |z| <= |x|
                        a, b = lhs.args
                        lb, ub = get_bounds(b)
                        if lb <= 0 <= ub:
                            raise ValueError("Attempting linearization of unsafe division, safen expression first (cpmpy/transformations/safen.py)")

                        r = intvar(*get_bounds(a % b)) # r is the remainder, reuse our bound calculations
                        mult_res, side_cons = get_or_make_var(b * rhs, csemap=csemap)
                        cpm_expr = eval_comparison(cpm_expr.name, a, mult_res + r)

                        # need absolute values of variables later
                        abs_of_a, side_cons_a = get_or_make_var(abs(a), csemap=csemap)
                        abs_of_b, side_cons_b = get_or_make_var(abs(b), csemap=csemap)
                        abs_of_rhs, side_cons_rhs = get_or_make_var(abs(rhs), csemap=csemap)
                        abs_of_r, side_cons_r = get_or_make_var(abs(r), csemap=csemap)
                        side_cons += side_cons_a + side_cons_b + side_cons_rhs + side_cons_r
                        # |r| < |b|
                        side_cons.append(abs_of_r < abs_of_b)

                        # ensure we round towards zero
                        mul_abs, extra_cons = get_or_make_var(abs_of_b * abs_of_rhs, csemap=csemap)
                        side_cons += extra_cons + [mul_abs <= abs_of_a]
                        newlist += linearize_constraint(side_cons, supported=supported, reified=reified, csemap=csemap)

                else:
                    raise TransformationNotImplementedError(f"lhs of constraint {cpm_expr} cannot be linearized, should"
                                                            f" be any of {supported | {'sub'} } but is {lhs}. "
                                                            f"Please report on github")

            elif isinstance(lhs, GlobalFunction) and lhs.name == "abs" and "abs" not in supported:
                if cpm_expr.name != "==": # TODO: remove this restriction, requires comparison flipping
                    newvar, newcons = get_or_make_var(lhs, csemap=csemap)
                    newlist += linearize_constraint(newcons, supported=supported, reified=reified, csemap=csemap)
                    cpm_expr = eval_comparison(cpm_expr.name, newvar, rhs)
                else:
                    x = lhs.args[0]
                    lb, ub = get_bounds(x)
                    if lb >= 0:  # always positive
                        newlist.append(x == rhs)
                    elif ub <= 0:  # always negative
                        newlist.append(x + rhs == 0)
                    else:
                        lhs_is_pos = cp.boolvar()
                        newcons = [lhs_is_pos.implies(x >= 0), (~lhs_is_pos).implies(x <= -1),
                                   lhs_is_pos.implies(x == rhs), (~lhs_is_pos).implies(x + rhs == 0)]
                        newlist += linearize_constraint(newcons, supported=supported, reified=reified, csemap=csemap)
                    continue # all should be linear now


            elif isinstance(lhs, GlobalFunction) and lhs.name not in supported:
                raise ValueError(f"Linearization of `lhs` ({lhs}) not supported, run "
                                 "`cpmpy.transformations.decompose_global.decompose_global() first")

            [cpm_expr] = canonical_comparison([cpm_expr])  # just transforms the constraint, not introducing new ones
            lhs, rhs = cpm_expr.args

            if lhs.name == "sum" and len(lhs.args) == 1 and isinstance(lhs.args[0], _BoolVarImpl) and "or" in supported:
                # very special case, avoid writing as sum of 1 argument
                new_expr = simplify_boolean([eval_comparison(cpm_expr.name,lhs.args[0], rhs)])
                assert len(new_expr) == 1
                newlist.append(Operator("or", new_expr))
                continue



            # check trivially true/false (not allowed by PySAT Card/PB)
            if cpm_expr.name in ('<', '<=', '>', '>=') and is_num(rhs):
                lb,ub = lhs.get_bounds()
                t_lb = eval_comparison(cpm_expr.name, lb, rhs)
                t_ub = eval_comparison(cpm_expr.name, ub, rhs)
                if t_lb and t_ub:
                    continue
                elif not t_lb and not t_ub:
                    newlist += linearize_constraint([BoolVal(False)], supported=supported, csemap=csemap) # post the linear version of False
                    break

            # now fix the comparisons themselves
            if cpm_expr.name == "<":
                new_rhs, cons = get_or_make_var(rhs - 1, csemap=csemap) # if rhs is constant, will return new constant
                newlist.append(lhs <= new_rhs)
                newlist += linearize_constraint(cons, csemap=csemap)
            elif cpm_expr.name == ">":
                new_rhs, cons = get_or_make_var(rhs + 1, csemap=csemap) # if rhs is constant, will return new constant
                newlist.append(lhs >= new_rhs)
                newlist += linearize_constraint(cons, csemap=csemap)
            elif cpm_expr.name == "!=":
                # Special case: BV != BV
                if isinstance(lhs, _BoolVarImpl) and isinstance(rhs, _BoolVarImpl):
                    newlist.append(lhs + rhs == 1)

                if reified or (isinstance(lhs, (Operator, GlobalConstraint)) and lhs.name not in {"sum","wsum"}):
                    # lhs is sum/wsum and rhs is constant OR
                    # lhs is GenExpr and rhs is constant or var
                    #  ... what requires less new variables?
                    # Big M implementation
                    # M is chosen so that
                    # lhs - rhs + 1 <= M*z
                    # rhs - lhs + 1 <= M*~z
                    # holds
                    z = boolvar()
                    # Calculate bounds of M = |lhs - rhs| + 1
                    _, M1 = (lhs - rhs + 1).get_bounds()
                    _, M2 = (rhs - lhs + 1).get_bounds()
                    cons = [lhs + -M1*z <= rhs-1, lhs  + -M2*z >= rhs-M2+1]
                    newlist += linearize_constraint(flatten_constraint(cons, csemap=csemap), supported=supported, reified=reified, csemap=csemap)

                else:
                    # introduce new indicator constraints
                    z = boolvar()
                    constraints = [z.implies(lhs < rhs), (~z).implies(lhs > rhs)]
                    newlist += linearize_constraint(constraints, supported=supported, reified=reified, csemap=csemap)
            else:
                # supported comparison
                newlist.append(eval_comparison(cpm_expr.name, lhs, rhs))

        elif cpm_expr.name == "alldifferent" and cpm_expr.name in supported:
            newlist.append(cpm_expr)
        elif cpm_expr.name == "alldifferent" and cpm_expr.name not in supported:
            """
                More efficient implementations possible
                http://yetanothermathprogrammingconsultant.blogspot.com/2016/05/all-different-and-mixed-integer.html
                Introduces n^2 new boolean variables
                Decomposes through bi-partite matching
            """
            # TODO check performance of implementation
            if reified is True:
                raise ValueError("Linear decomposition of AllDifferent does not work reified. "
                                 "Ensure 'alldifferent' is not in the 'supported_nested' set of 'decompose_in_tree'")

            lbs, ubs = get_bounds(cpm_expr.args)
            lb, ub = min(lbs), max(ubs)
            n_vals = (ub-lb) + 1

            x = boolvar(shape=(len(cpm_expr.args), n_vals))

            newlist += [sum(row) == 1 for row in x]   # each var has exactly one value
            newlist += [sum(col) <= 1 for col in x.T] # each value can be taken at most once

            # link Boolean matrix and integer variable
            for arg, row in zip(cpm_expr.args, x):
                if is_num(arg): # constant, fix directly
                    newlist.append(Operator("sum", [row[arg-lb]]) == 1) # ensure it is linear
                else: # ensure result is canonical
                    newlist.append(sum(np.arange(lb, ub + 1) * row) + -1 * arg == 0)

        elif isinstance(cpm_expr, (DirectConstraint, BoolVal)):
            newlist.append(cpm_expr)

        elif isinstance(cpm_expr, GlobalConstraint) and cpm_expr.name not in supported:
            raise ValueError(f"Linearization of global constraint {cpm_expr} not supported, run "
                             f"`cpmpy.transformations.decompose_global.decompose_global() first")

    return newlist

def only_positive_bv(lst_of_expr, csemap=None):
    """
        Replaces :class:`~cpmpy.expressions.comparison.Comparison` containing :class:`~cpmpy.expressions.variables.NegBoolView` with equivalent expression using only :class:`~cpmpy.expressions.variables.BoolVar`.
        Comparisons are expected to be linearized. Only apply after applying :func:`linearize_constraint(cpm_expr) <linearize_constraint>`.

        Resulting expression is linear if the original expression was linear.
    """
    newlist = []
    for cpm_expr in lst_of_expr:

        if isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            new_lhs = lhs
            new_cons = []

            if isinstance(lhs, _NumVarImpl) or lhs.name in {"sum","wsum"}:
                new_lhs, const = only_positive_bv_wsum_const(lhs)
                rhs -= const
            else:
                # other operators in comparison such as "min", "max"
                nbv_sel = [isinstance(a, NegBoolView) for a in lhs.args]
                if any(nbv_sel):
                    new_args = []
                    for i, nbv in enumerate(nbv_sel):
                        if nbv:
                            aux = cp.boolvar()
                            new_args.append(aux)
                            new_cons += [aux + lhs.args[i]._bv == 1]  # aux == 1 - arg._bv
                        else:
                            new_args.append(lhs.args[i])

                    new_lhs = copy.copy(lhs)
                    new_lhs.update_args(new_args)

            if new_lhs is not lhs:
                newlist.append(eval_comparison(cpm_expr.name, new_lhs, rhs))
                newlist += new_cons  # already linear
            else:
                newlist.append(cpm_expr)

        # reification
        elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            cond, subexpr = cpm_expr.args
            assert isinstance(cond, _BoolVarImpl), f"{cpm_expr} is not a supported linear expression. Apply " \
                                                   f"`linearize_constraint` before calling `only_positive_bv` "
            # BV -> Expr
            subexpr = only_positive_bv([subexpr], csemap=csemap)
            newlist += [cond.implies(expr) for expr in subexpr]

        elif isinstance(cpm_expr, _BoolVarImpl):
            raise ValueError(f"Unreachable: unexpected Boolean literal (`_BoolVarImpl`) in expression {cpm_expr}, perhaps `linearize_constraint` was not called before this `only_positive_bv `call")
        elif isinstance(cpm_expr, (GlobalConstraint, BoolVal, DirectConstraint)):
            newlist.append(cpm_expr)
        else:
            raise Exception(f"{cpm_expr} is not linear or is not supported. Please report on github")

    return newlist

def only_positive_bv_wsum(expr):
    """
        Replaces a var/sum/wsum expression containing :class:`~cpmpy.expressions.variables.NegBoolView` with an equivalent expression 
        using only :class:`~cpmpy.expressions.variables.BoolVar`. 

        It might add a constant term to the expression, if you want the constant separately, use :func:`only_positive_bv_wsum_const`.
        
        Arguments:
        - `cpm_expr`: linear expression (sum, wsum, var)
        
        Returns tuple of:
        - `pos_expr`: linear expression (sum, wsum, var) without NegBoolView
    """
    if isinstance(expr, _NumVarImpl) or expr.name in {"sum","wsum"}:
        pos_expr, const = only_positive_bv_wsum_const(expr)
        if const == 0:
            return pos_expr
        else:
            assert isinstance(pos_expr, Operator) and pos_expr.name == "wsum", f"unexpected expression, should be wsum but got {pos_expr}"
            # should we check if it already has a constant term?
            return Operator("wsum", [pos_expr.args[0]+[1], pos_expr.args[1]+[const]])
    else:
        return expr

def only_positive_bv_wsum_const(cpm_expr):
    """
        Replaces a var/sum/wsum expression containing :class:`~cpmpy.expressions.variables.NegBoolView` with an equivalent expression 
        using only :class:`~cpmpy.expressions.variables.BoolVar` as well as a constant term that must be added to the new expression to be equivalent.

        If you want the expression where the constant term is part of the wsum returned, use :func:`only_positive_bv_wsum`.
        
        Arguments:
        - `cpm_expr`: linear expression (sum, wsum, var)
        
        Returns tuple of:
        - `pos_expr`: linear expression (sum, wsum, var) without NegBoolView
        - `const`: The difference between the original expression and the new expression, 
                   i.e. a constant term that must be added to pos_expr to be an equivalent linear expression.
    """
    if isinstance(cpm_expr, _NumVarImpl):
        if isinstance(cpm_expr,NegBoolView):
            return Operator("wsum",[[-1], [cpm_expr._bv]]), 1
        else:
            return cpm_expr, 0

    elif cpm_expr.name == "sum":
        # indicator on arguments being negboolviews
        nbv_sel = [isinstance(a, NegBoolView) for a in cpm_expr.args]
        if any(nbv_sel):
            const = 0
            weights = []
            variables = []
            for i, nbv in enumerate(nbv_sel):
                if nbv:
                    const += 1
                    weights.append(-1)
                    variables.append(cpm_expr.args[i]._bv)
                else:
                    weights.append(1)
                    variables.append(cpm_expr.args[i])
            return Operator("wsum", [weights, variables]), const
        else:
            return cpm_expr, 0

    elif cpm_expr.name == "wsum":
        # indicator on arguments of the wsum variable being negboolviews
        nbv_sel = [isinstance(a, NegBoolView) for a in cpm_expr.args[1]]
        if any(nbv_sel):
            # copy weights and variables
            weights = [w for w in cpm_expr.args[0]]
            variables = [v for v in cpm_expr.args[1]]
            const = 0
            for i, nbv in enumerate(nbv_sel):
                if nbv:
                    const += weights[i]
                    weights[i] = -weights[i]
                    variables[i] = variables[i]._bv
            return Operator("wsum", [weights, variables]), const
        else:
            return cpm_expr, 0

    else:
        raise ValueError(f"unexpected expression, should be sum, wsum or var but got {cpm_expr}")


def canonical_comparison(lst_of_expr):
    """
        Canonicalize a comparison expression.
        Transforms linear expressions, or a reification thereof into canonical form by:
            - moving all variables to the left-hand side
            - moving constants to the right-hand side

        Expects the input constraints to be flat. Only apply after applying :func:`flatten_constraint`
    """

    lst_of_expr = toplevel_list(lst_of_expr) # ensure it is a list

    newlist = []
    for cpm_expr in lst_of_expr:

        if isinstance(cpm_expr, Operator) and cpm_expr.name == '->':    # half reification of comparison
            lhs, rhs = cpm_expr.args
            if isinstance(rhs, Comparison):
                rhs = canonical_comparison(rhs)[0]
                newlist.append(lhs.implies(rhs))
            elif isinstance(lhs, Comparison):
                lhs = canonical_comparison(lhs)[0]
                newlist.append(lhs.implies(rhs))
            else:
                newlist.append(cpm_expr)

        elif isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            if isinstance(lhs, Comparison) and (is_bool(rhs) or isinstance(rhs, Expression) and rhs.is_bool()):
                assert cpm_expr.name == "==", "Expected a reification of a comparison here, but got {}".format(cpm_expr.name)
                lhs = canonical_comparison(lhs)[0]
            elif is_num(lhs) or isinstance(lhs, _NumVarImpl) or (isinstance(lhs, Operator) and lhs.name in {"sum", "wsum", "sub"}):
                if lhs.name == "sub":
                    lhs = Operator("wsum", [[1,-1],lhs.args])
                # bring all vars to lhs
                lhs2 = []
                if isinstance(rhs, _NumVarImpl):
                    lhs2, rhs = [-1 * rhs], 0
                elif isinstance(rhs, BoolVal):
                    lhs2, rhs = [-1] if rhs.value() else [], 0 
                elif isinstance(rhs, Operator) and rhs.name == "-":
                    lhs2, rhs = [rhs.args[0]], 0
                elif isinstance(rhs, Operator) and rhs.name == "sum":
                    lhs2, rhs = [-1 * b if isinstance(b, _NumVarImpl) else -1 * b.args[0] for b in rhs.args
                                 if isinstance(b, _NumVarImpl) or isinstance(b, Operator)], \
                                 sum(b for b in rhs.args if is_num(b))
                elif isinstance(rhs, Operator) and rhs.name == "wsum":
                    lhs2, rhs = [-a * b for a, b in zip(rhs.args[0], rhs.args[1])
                                    if isinstance(b, _NumVarImpl)], \
                                    sum(-a * b for a, b in zip(rhs.args[0], rhs.args[1])
                                    if not isinstance(b, _NumVarImpl))
                
                # 2) add collected variables to lhs
                if isinstance(lhs, Operator) and lhs.name == "sum":
                    lhs = sum([1 * a for a in lhs.args] + lhs2)
                elif isinstance(lhs, _NumVarImpl) or (isinstance(lhs, Operator) and lhs.name == "wsum"):
                    lhs = lhs + lhs2
                else:
                    raise ValueError(f"unexpected expression on lhs of expression, should be sum, wsum or intvar but got {lhs}")

                assert not is_num(lhs), "lhs cannot be an integer at this point!"

                # bring all const to rhs
                if isinstance(lhs, Operator):
                    if lhs.name == "sum":
                        new_args = []
                        for i, arg in enumerate(lhs.args):
                            if is_num(arg):
                                rhs -= arg
                            else:
                                new_args.append(arg)
                        lhs = Operator("sum", new_args)

                    elif lhs.name == "wsum":
                        new_weights, new_args = [], []
                        for i, (w, arg) in enumerate(zip(*lhs.args)):
                            if is_num(arg):
                                rhs -= w * arg
                            else:
                                new_weights.append(w)
                                new_args.append(arg)
                        lhs = Operator("wsum", [new_weights, new_args])
                    else:
                        raise ValueError(f"lhs should be sum or wsum, but got {lhs}")
                else:
                    assert isinstance(lhs, _NumVarImpl), f"Expected variable here, but got {lhs} in expression {cpm_expr}"
                    lhs = Operator("sum", [lhs])

            newlist.append(eval_comparison(cpm_expr.name, lhs, rhs))
        else:   # rest of expressions
            newlist.append(cpm_expr)

    return newlist

def only_positive_coefficients(lst_of_expr):
    """
        Replaces Boolean terms with negative coefficients in linear constraints with terms with positive coefficients by negating its literal.
        This can simplify a `wsum` into `sum`.
        `cpm_expr` is expected to be a canonical comparison.
        Only apply after applying :func:`canonical_comparison(cpm_expr) <canonical_comparison>`

        Resulting expression is linear.
    """
    newlist = []
    for cpm_expr in lst_of_expr:
        if isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args

            #    ... -c*b + ... <= k
            # :: ... -c*(1 - ~b) + ... <= k
            # :: ... -c + c* ~b + ... <= k
            # :: ... + c*~b + ... <= k+c
            if lhs.name == "wsum":
                weights, args = lhs.args
                idxes = {i for i, (w, a) in enumerate(zip(weights, args)) if w < 0 and isinstance(a, _BoolVarImpl)}
                nw, na = zip(*[(-w, ~a) if i in idxes else (w, a) for i, (w, a) in enumerate(zip(weights, args))])
                rhs += sum(-weights[i] for i in idxes)

                # Simplify wsum to sum if all weights are 1
                if all(w == 1 for w in nw):
                    lhs = Operator("sum", [list(na)])
                else:
                    lhs = Operator("wsum", [list(nw), list(na)])

            newlist.append(eval_comparison(cpm_expr.name, lhs, rhs))

        # reification
        elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            cond, subexpr = cpm_expr.args
            assert isinstance(cond, _BoolVarImpl), f"{cpm_expr} is not a supported linear expression. Apply " \
                                                   f"`linearize_constraint` before calling `only_positive_coefficients` "
            subexpr = only_positive_coefficients([subexpr])
            newlist += [cond.implies(expr) for expr in subexpr]

        else:
            newlist.append(cpm_expr)

    return newlist



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## exact.py
##
"""
    Interface to Exact's Python API

    Exact solves decision and optimization problems formulated as integer linear programs. 
    Under the hood, it converts integer variables to binary (0-1) variables and applies highly efficient 
    propagation routines and strong cutting-planes / pseudo-Boolean conflict analysis.

    The solver's git repository:
    https://gitlab.com/nonfiction-software/exact

    Always use :func:`cp.SolverLookup.get("exact") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'exact' python package is installed:

    .. code-block:: console
    
        $ pip install exact

    .. warning::
        Exact requires Python 3.10 or higher and the pip install only works on Linux and Windows.
        On MacOS, you have to install the package from source.

    See https://pypi.org/project/exact for more information.

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_exact

    ==============
    Module details
    ==============
"""
import sys  # for stdout checking
import time
from typing import Optional

import pkg_resources
from pkg_resources import VersionConflict

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import *
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl
from ..transformations.comparison import only_numexpr_equality
from ..transformations.flatten_model import flatten_constraint, flatten_objective
from ..transformations.get_variables import get_variables
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.linearize import linearize_constraint, only_positive_bv, only_positive_bv_wsum
from ..transformations.reification import only_implies, reify_rewrite, only_bv_reifies
from ..transformations.normalize import toplevel_list
from ..transformations.safening import no_partial_functions
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.utils import flatlist, argvals, argval
from ..exceptions import NotSupportedError

import numpy as np
import numbers

class CPM_exact(SolverInterface):
    """
    Interface to Exact's Python API

    Creates the following attributes (see parent constructor for more):

    - ``xct_solver`` : the Exact instance used in solve() and solveAll()
    - ``assumption_dict`` : maps Exact variables to (Exact value, CPM assumption expression)

    Documentation of the solver's own Python API is sparse, but example usage can be found at:
    https://gitlab.com/nonfiction-software/exact/-/tree/main/python_examples
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            # check if exact is installed
            import exact
            # check installed version
            pkg_resources.require("exact>=2.1.0")
            return True
        except ModuleNotFoundError: # exact is not installed
            return False
        except VersionConflict: # unsupported version of exact
            warnings.warn(f"CPMpy requires Exact version >=2.1.0 is required but you have version "
                          f"{pkg_resources.get_distribution('exact').version}, beware exact>=2.1.0 requires "
                          f"Python 3.10 or higher.")
            return False
        except Exception as e:
            raise e
        
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('exact').version
        except pkg_resources.DistributionNotFound:
            return None


    def __init__(self, cpm_model=None, subsolver=None, **kwargs):
        """
        Constructor of the native solver object

        Requires a CPMpy model as input, and will create the corresponding
        Exact solver object xct_solver.

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: None

        Exact takes options at initialization instead of solving.
        The Exact solver parameters are defined by https://gitlab.com/nonfiction-software/exact/-/blob/main/src/Options.hpp
        Note some parameters use "-" which cannot be used in a keyword argument.
        A workaround is to use dict-unpacking: `CPM_Exact(**{parameter-with-hyphen: 42})`
        """
        if not self.supported():
            raise Exception("CPM_exact: Install the python package 'exact' to use this solver interface.")
        
        assert subsolver is None, "Exact does not allow subsolvers."

        from exact import Exact as xct
        # initialise the native solver object
        options = list(kwargs.items()) # options is a list of string-pairs, e.g. [("verbosity","1")]
        options = [(opt[0], str(opt[1])) for opt in options] # Ensure values are also strings
        self.xct_solver = xct(options)

        # can override encoding of variables
        self.encoding = None

        # for solving with assumption variables,
        self.assumption_dict = None
        self.objective_ = None
        self.objective_is_min_ = True

        # initialise everything else and post the constraints/objective
        super().__init__(name="exact", cpm_model=cpm_model)
        
    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.xct_solver

    def _fillVars(self):
        if not self.xct_solver.hasSolution():
            self.objective_value_ = None
            for cpm_var in self.user_vars:
                cpm_var._value = None
            return

        # fill in variable values
        lst_vars = list(self.user_vars)
        exact_vals = self.xct_solver.getLastSolutionFor(self.solver_vars(lst_vars))
        for cpm_var, val in zip(lst_vars,exact_vals):
            cpm_var._value = bool(val) if isinstance(cpm_var, _BoolVarImpl) else val # xct value is always an int

    def solve(self, time_limit=None, assumptions=None, **kwargs):
        """
            Call Exact

            Overwrites ``self.cpm_status``

            :param assumptions: CPMpy Boolean variables (or their negation) that are assumed to be true.
                           For repeated solving, and/or for use with :func:`s.get_core() <get_core()>`: if the model is UNSAT,
                           get_core() returns a small subset of assumption variables that are unsat together.
            :type assumptions: list of CPMpy Boolean variables

            :param time_limit: optional, time limit in seconds
            :type time_limit: int or float

            :return: Bool:
                - True      if a solution is found (not necessarily optimal, e.g. could be after timeout)
                - False     if no solution is found
        """
        from exact import Exact as xct
 
        # set additional keyword arguments
        if(len(kwargs.items())>0):
            wrn_txt = f"Exact only supports options at initialization. Ignoring additional options {kwargs.items()}\n"
            wrn_txt += "Use cp.SolverLookup.lookup('exact', **{parameter-with-hyphen: 42}) to set Exact parameters"
            warnings.warn(wrn_txt)

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        self.xct_solver.clearAssumptions()

        # set assumptions
        if assumptions is not None:
            assert all(v.is_bool() for v in assumptions), "Non-Boolean assumptions given to Exact: " + str([v for v in assumptions if not v.is_bool()])
            assump_vals = [int(not isinstance(v, NegBoolView)) for v in assumptions]
            assump_vars = [self.solver_var(v._bv if isinstance(v, NegBoolView) else v) for v in assumptions]
            self.assumption_dict = {xct_var: (xct_val,cpm_assump) for (xct_var, xct_val, cpm_assump) in zip(assump_vars,assump_vals,assumptions)}
            self.xct_solver.setAssumptions(list(zip(assump_vars,assump_vals)))

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            timeout = time_limit
        else:
            timeout = 0
            
        # call the solver, with parameters
        start = time.time()
        my_status, obj_val = self.xct_solver.toOptimum(timeout=timeout)
        end = time.time()

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = end - start

        self.objective_value_ = None
        # translate exit status
        #   see 'toOptimum' documentation:
        #   https://gitlab.com/nonfiction-software/exact/-/blob/main/src/interface/IntProg.cpp#L877
        if my_status == "UNSAT": # found unsatisfiability
            if self.has_objective() and self.xct_solver.hasSolution(): # optimisation problem -> unsat = no better solution found
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else:
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif my_status == "SAT": # the optimal value has been found
            assert self.xct_solver.hasSolution()
            # COP
            if self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            # CSP
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif my_status == "INCONSISTENT": # found inconsistency over assumptions
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif my_status == "TIMEOUT": # found timeout
            if self.xct_solver.hasSolution(): # found a (sub-)optimal solution
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            else: # no solution found
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:
            raise NotImplementedError(my_status)  # a new status type was introduced, please report on github
        
        self._fillVars()
        if self.has_objective():
            if self.objective_is_min_:
                self.objective_value_ = obj_val
            else: # maximize, so actually negative value
                self.objective_value_ = -obj_val
        
        # True/False depending on self.cpm_status
        return self._solve_return(self.cpm_status)

    def _update_time(self, timelim, start, end):
        """
            Internal helper function to keep track of remaining time.
        """
        if timelim != 0:
            timelim -= (end - start)
            if timelim == 0: timelim = -1
        return timelim

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            Compute all solutions and optionally, display the solutions.

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                time_limit: stop after this many seconds (default: None)
                solution_limit: stop after this many solutions (default: None)
                call_from_model: whether the method is called from a CPMpy Model instance or not
                any other keyword argument

            Returns: number of solutions found
        """
        # set additional keyword arguments
        if len(kwargs.items()) > 0:
            warnings.warn(f"Exact only supports options at initialization: {kwargs.items()}")

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        self.xct_solver.clearAssumptions()

        timelim = time_limit if time_limit is not None else 0

        total_start = time.time()

        if self.has_objective():
            start = time.time()
            if not call_from_model:
                warnings.warn("Adding constraints to solver object to find all solutions, solver state will be invalid after this call!")

            (my_status, objval) = self.xct_solver.toOptimum(timelim) # fix the solution to the optimal objective
            if my_status == "UNSAT": # found unsatisfiability
                total_end = time.time()
                self._fillVars() # erases the solution
                # update exit status
                self.cpm_status = SolverStatus(self.name)
                self.cpm_status.runtime = total_end - total_start
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
                # early exit
                return 0
            elif my_status == "INCONSISTENT": # found inconsistency
                raise ValueError("Error: inconsistency during solveAll should not happen, please warn the developers of this bug")
            elif my_status == "TIMEOUT": # found timeout
                total_end = time.time()
                # update exit status
                self.cpm_status = SolverStatus(self.name)
                self.cpm_status.runtime = total_end - total_start
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
                # early exit
                return 0
            else:
                assert my_status == "SAT", "Unexpected status from Exact"
            self += self.objective_ == objval # fix obj val
            end = time.time()
            timelim = self._update_time(timelim, start, end) # update remaining time


        solsfound = 0
        while timelim >= 0 and (solution_limit is None or solsfound < solution_limit):
            # call the solver, with parameters
            start = time.time()
            my_status = self.xct_solver.runFull(optimize=False, timeout=timelim)
            end = time.time()
            timelim = self._update_time(timelim, start, end) # update remaining time

            assert my_status in ["UNSAT","SAT","INCONSISTENT","TIMEOUT"], "Unexpected status code for Exact: " + my_status
            if my_status == "UNSAT": # found unsatisfiability (no more solutions to be found)
                self._fillVars() # erases the solution
                break
            elif my_status == "SAT": # found solution, but not optimality proven
                assert self.xct_solver.hasSolution()
                solsfound += 1
                self.xct_solver.invalidateLastSol() # TODO: pass user vars to this function
                if display is not None:
                    self._fillVars()
                    if isinstance(display, Expression):
                        print(argval(display))
                    elif isinstance(display, list):
                        print(argvals(display))
                    else:
                        display()  # callback
            elif my_status == "INCONSISTENT": # found inconsistency
                raise ValueError("Error: inconsistency during solveAll should not happen, please warn the developers of this bug")
            elif my_status == "TIMEOUT": # found timeout
                break
        total_end = time.time()

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = total_end - total_start

        if solsfound: # found some solutions
            if solsfound == solution_limit: # matched solution limit
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif my_status == "UNSAT": # found all solutions (before limits were reached)
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else: # timeout
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        else:
            if my_status == "UNSAT": # unsat problem
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            elif my_status == "TIMEOUT": # timeout
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                raise NotImplementedError(my_status) # please report on GitHub

        return solsfound

    def has_objective(self):
        """
            Returns whether the solver has an objective function or not.
        """
        return self.objective_ is not None


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var):  # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view. Should be eliminated in linearize
        if isinstance(cpm_var, NegBoolView):
            raise NotSupportedError("Negative literals should not be left as part of any equation. Please report.")

        # return it if it already exists
        if cpm_var in self._varmap:
            return self._varmap[cpm_var]

        # create if it does not exist
        revar = str(cpm_var)
        if isinstance(cpm_var, _BoolVarImpl):
            self.xct_solver.addVariable(revar)
        elif isinstance(cpm_var, _IntVarImpl):
            lb, ub = cpm_var.get_bounds()
            if self.encoding is None:
                encoding = "order" if ub-lb < 8 else "log" # heuristic bound
            else:
                encoding = self.encoding # can also force it
            self.xct_solver.addVariable(revar, lb, ub, encoding)
        else:
            raise NotImplementedError("Not a known var {}".format(cpm_var))
        self._varmap[cpm_var] = revar
        return revar


    def objective(self, expr, minimize):
        """
            Post the given expression to the solver as objective to minimize/maximize.

            Arguments:
                expr: Expression, the CPMpy expression that represents the objective function
                minimize: Bool, whether it is a minimization problem (True) or maximization problem (False)

        """
        self.objective_ = expr
        self.objective_is_min_ = minimize

        # make objective function non-nested and with positive BoolVars only
        (flat_obj, flat_cons) = flatten_objective(expr, csemap=self._csemap)
        flat_obj = only_positive_bv_wsum(flat_obj)  # remove negboolviews
        self.user_vars.update(get_variables(flat_obj))  # add objvars to vars
        self += flat_cons  # add potentially created constraints

        # make objective function or variable and post
        xct_cfvars,xct_rhs = self._make_numexpr(flat_obj,0)

        # TODO: make this a custom transformation?
        newcfvrs = []
        for c,v in xct_cfvars:
            if is_num(v):
                xct_rhs += c*v
            else:
                newcfvrs.append((c,v))

        self.xct_solver.setObjective(newcfvrs,minimize,xct_rhs)

    @staticmethod
    def fix(o):
        x = o.item() if isinstance(o, np.generic) else o
        if not isinstance(x, numbers.Integral):
            raise NotImplementedError("Exact requires all values to be integral")
        return x

    def _make_numexpr(self, lhs, rhs):
        """
            Turns a numeric CPMpy 'flat' expression into a solver-specific
            numeric expression
        """

        xcfvars = []
        xrhs = 0
        
        assert is_num(rhs), "RHS of inequality should be numeric after transformations: {}".format(rhs)
        xrhs += rhs

        if is_num(lhs):
            xrhs -= lhs
        elif isinstance(lhs, _NumVarImpl):
            xcfvars = [(1,self.solver_var(lhs))]
        elif lhs.name == "sum":
            xcfvars = [(1,x) for x in self.solver_vars(lhs.args)]
        elif lhs.name == "wsum":
            xcfvars = list(zip([self.fix(c) for c in lhs.args[0]],self.solver_vars(lhs.args[1])))
        else:
            raise NotImplementedError("Exact: Unexpected lhs {} for expression {}".format(lhs.name,lhs))

        return xcfvars, self.fix(xrhs)


    def transform(self, cpm_expr):
        """
        Transform arbitrary CPMpy expressions to constraints the solver supports

        Implemented through chaining multiple solver-independent **transformation functions** from
        the `cpmpy/transformations/` directory.

        See the :ref:`Adding a new solver` docs on readthedocs for more information.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: list of Expression
        """

        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel={"mod", "div"}) # linearize expects safe exprs
        cpm_cons = decompose_in_tree(cpm_cons, supported=frozenset({'alldifferent', 'abs'}), csemap=self._csemap) # Abs and Alldiff have a specialized MIP decomp
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['sum', 'wsum']), csemap=self._csemap)  # constraints that support reification
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum"]), csemap=self._csemap)  # supports >, <, !=
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)  # anything that can create full reif should go above...
        cpm_cons = linearize_constraint(cpm_cons, supported=frozenset({"sum","wsum","mul"}), csemap=self._csemap)  # the core of the MIP-linearization
        cpm_cons = only_positive_bv(cpm_cons, csemap=self._csemap)  # after linearisation, rewrite ~bv into 1-bv

        return cpm_cons

        # NOTE: the transformations that are still done specifically for Exact are two-fold:
        # transform '==' and '<=' to '>='
        #
        # this seems quite general and is a candidate to function as an independent transformation.

    def _add_xct_constr(self, xct_cfvars, uselb, lb, useub, ub):
        self.xct_solver.addConstraint(xct_cfvars, uselb, self.fix(lb), useub, self.fix(ub))

    def _add_xct_reif_right(self, head, sign, xct_cfvars, xct_rhs):
        self.xct_solver.addRightReification(head, sign, xct_cfvars, self.fix(xct_rhs))

    @staticmethod
    def is_multiplication(cpm_expr): # helper function
        return isinstance(cpm_expr, Operator) and cpm_expr.name == 'mul'

    def add(self, cpm_expr_orig):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr_orig: CPMpy expression, or list thereof
        :type cpm_expr_orig: Expression or list of Expression

        :return: self
        """

        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        # transform and post the constraints
        for cpm_expr in self.transform(cpm_expr_orig):
            # Comparisons: only numeric ones as 'only_implies()' has removed the '==' reification for Boolean expressions
            # numexpr `comp` bvar|const
            if isinstance(cpm_expr, Comparison):
                lhs, rhs = cpm_expr.args
                if cpm_expr.name == "==":
                    assert isinstance(lhs, Operator)
                    # can be sum, wsum or mul
                    if lhs.name == "mul":
                        assert pkg_resources.require("exact>=2.1.0"), f"Multiplication constraint {cpm_expr} " \
                                                                      f"only supported by Exact version 2.1.0 and above"
                        if is_num(rhs): # make dummy var
                            rhs = cp.intvar(rhs, rhs)
                        xct_rhs = self.solver_var(rhs)
                        assert all(isinstance(v, _IntVarImpl) for v in lhs.args), "constant * var should be " \
                                                                                  "rewritten by linearize"
                        self.xct_solver.addMultiplication(self.solver_vars(lhs.args), True, xct_rhs, True, xct_rhs)

                    else:
                        assert lhs.name == "sum" or lhs.name == "wsum"
                        xct_cfvars, xct_rhs = self._make_numexpr(lhs, rhs)
                        self._add_xct_constr(xct_cfvars, True, xct_rhs, True, xct_rhs)

                elif cpm_expr.name == "<=":
                    xct_cfvars, xct_rhs = self._make_numexpr(lhs, rhs)
                    self._add_xct_constr(xct_cfvars, False, 0, True, xct_rhs)

                elif cpm_expr.name == ">=":
                    xct_cfvars, xct_rhs = self._make_numexpr(lhs, rhs)
                    self._add_xct_constr(xct_cfvars, True, xct_rhs, False, 0)

                else:
                    raise NotImplementedError(cpm_expr)  # if you reach this... please report on github

            elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
                # Indicator constraints
                # Take form bvar -> sum(x,y,z) >= rvar
                cond, sub_expr = cpm_expr.args
                assert isinstance(cond, _BoolVarImpl), f"Implication constraint {cpm_expr} must have BoolVar as lhs"
                assert isinstance(sub_expr, Comparison), "Implication must have linear constraints on right hand side"

                if sub_expr.name not in ["==", ">=", "<="]:
                    raise NotImplementedError("Constraint not supported by Exact '{}' {}".format(lhs.name, cpm_expr))

                lhs, rhs = sub_expr.args
                xct_cfvars, xct_rhs = self._make_numexpr(lhs,rhs)

                if isinstance(cond, NegBoolView):
                    cond, bool_val = self.solver_var(cond._bv), False
                else:
                    cond, bool_val = self.solver_var(cond), True

                if sub_expr.name in ["==", ">="]:
                    # a -> (b >=c)
                    self._add_xct_reif_right(cond, bool_val, xct_cfvars, xct_rhs)
                if sub_expr.name in ["==", "<="]:
                    # a -> (b =< c), rewrite to a -> (-b >= -c)
                    self._add_xct_reif_right(cond, bool_val, [(-x,y) for x,y in xct_cfvars], -xct_rhs)

            # True or False
            elif isinstance(cpm_expr, BoolVal):
                self._add_xct_constr([], True, 0 if cpm_expr.args[0] else 1, False, 0)

            # a direct constraint, pass to solver
            elif isinstance(cpm_expr, DirectConstraint):
                cpm_expr.callSolver(self, self.xct_solver)
                return self

            else:
                raise NotImplementedError(cpm_expr)  # if you reach this... please report on github
            
        return self
    __add__ = add  # avoid redirect in superclass

    def get_core(self):
        from exact import Exact as xct
        """
            For use with s.solve(assumptions=[...]). Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those variables that are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal, though this interface does open up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!
        """
        assert self.assumption_dict is not None,  "get_core(): requires a list of assumption variables, e.g. s.solve(assumptions=[...])"

        # return cpm_variables corresponding to Exact core
        return [self.assumption_dict[i][1] for i in self.xct_solver.getLastCore()]


    def solution_hint(self, cpm_vars, vals):
        """
        Exact supports warmstarting the solver with a partial feasible assignment.

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        # clear previous solution hints
        self.xct_solver.clearSolutionHints(self.solver_vars(list(self.user_vars)))

        cpm_vars = flatlist(cpm_vars)
        vals = flatlist(vals)
        assert (len(cpm_vars) == len(vals)), "Variables and values must have the same size for hinting"
        self.xct_solver.setSolutionHints(list(zip(self.solver_vars(cpm_vars), vals)))



"""
    CLI script for running XCSP3 instances.

    A command-line interface for solving XCSP3 constraint satisfaction problems using CPMpy.

    Example usage:

    .. code-block:: console

        python xcsp3_cpmpy.py <xcsp3_instance.xml>


    Required Arguments
    ------------------
    benchname : str
        Path to the XCSP3 XML file to solve.

    Optional XCSP3 Arguments
    ------------------------
    -s, --seed : int, optional
        Random seed (between 0 and 4294967295) for reproducibility.
    -l, --time-limit : int, optional
        CPU time limit in seconds. The process will be killed if this limit is exceeded.
    -m, --mem-limit : int, optional
        Memory usage limit in MiB (1 MiB = 1024 * 1024 bytes).
    -t, --tmpdir : str, optional
        Directory where temporary read/write operations are allowed.
    -c, --cores : int, optional
        Number of processing units to use (logical cores or processors).

    Required CPMpy Argument
    -----------------------
    --solver : str
        The name of the CPMpy solver to use. Can be in the format "solver" or "solver:subsolver".

    Optional CPMpy Arguments
    ------------------------
    --time-buffer : int, optional
        Time in seconds reserved before hitting the time limit, used for solver cleanup and solution printing.
    --intermediate : bool, optional
        If specified, intermediate solutions will be printed (if supported by the solver).
"""

from __future__ import annotations

import argparse
import lzma
import warnings
import psutil
import signal
import time
import sys, os
import random


if sys.platform != "win32":
    import resource
    
import pathlib
from pathlib import Path
from enum import Enum
from typing import Optional
from io import StringIO 

from contextlib import contextmanager


# CPMpy
import cpmpy as cp
from cpmpy.solvers.solver_interface import ExitStatus as CPMStatus
from cpmpy.tools.xcsp3 import _parse_xcsp3, _load_xcsp3
from cpmpy.tools.xcsp3 import globals
from cpmpy.solvers.ortools import CPM_ortools
from cpmpy.expressions.variables import _BV_PREFIX, _IV_PREFIX

# PyCSP3
from xml.etree.ElementTree import ParseError
import xml.etree.cElementTree as ET

# Utils
import os, pathlib
sys.path.append(os.path.join(pathlib.Path(__file__).parent.resolve()))


# Configuration
SUPPORTED_SOLVERS = ["choco", "ortools", "exact", "z3", "minizinc", "gurobi"]
SUPPORTED_SUBSOLVERS = {
    "minizinc": ["gecode", "chuffed"]
}
DEFAULT_SOLVER = "ortools"
TIME_BUFFER = 5 # seconds
# TODO : see if good value
MEMORY_BUFFER_SOFT = 2 # MiB
MEMORY_BUFFER_HARD = 0 # MiB
MEMORY_BUFFER_SOLVER = 20 # MB

original_stdout = sys.stdout

def sigterm_handler(_signo, _stack_frame):
    """
        Handles a SIGTERM. Gives us 1 second to finish the current job before we get killed.
    """
    # Report that we haven't found a solution in time
    print_status(ExitStatus.unknown)
    print_comment("SIGTERM raised.")
    sys.exit(0)
    
def rlimit_cpu_handler(_signo, _stack_frame):
    """
        Handles a SIGXCPU.
    """
    # Report that we haven't found a solution in time
    print_status(ExitStatus.unknown)
    print_comment("SIGXCPU raised.")
    print(flush=True)
    sys.exit(0)

def init_signal_handlers():
    """
    Configure signal handlers
    """
    signal.signal(signal.SIGINT, sigterm_handler)
    signal.signal(signal.SIGTERM, sigterm_handler)
    signal.signal(signal.SIGINT, sigterm_handler)
    signal.signal(signal.SIGABRT, sigterm_handler)
    if sys.platform != "win32":
        signal.signal(signal.SIGXCPU, rlimit_cpu_handler)
    else:
        warnings.warn("Windows does not support setting SIGXCPU signal")

def set_memory_limit(mem_limit, verbose:bool=False):
    """
    Set memory limit (Virtual Memory Size). 
    """
    if mem_limit is not None:
        soft = max(mib_as_bytes(mem_limit) - mib_as_bytes(MEMORY_BUFFER_SOFT), mib_as_bytes(MEMORY_BUFFER_SOFT))
        hard = max(mib_as_bytes(mem_limit) - mib_as_bytes(MEMORY_BUFFER_HARD), mib_as_bytes(MEMORY_BUFFER_HARD))
        if verbose:
            print_comment(f"Setting memory limit: {soft} -- {hard}")
        if sys.platform != "win32":
            resource.setrlimit(resource.RLIMIT_AS, (soft, hard)) # limit memory in number of bytes
        else:
            warnings.warn("Memory limits using `resource` are not supported on Windows. Skipping hard limit.")

def set_time_limit(time_limit, verbose:bool=False):
    """
    Set time limit (CPU time in seconds).
    """
    if time_limit is not None:
        if sys.platform != "win32":
            soft = time_limit
            hard = resource.RLIM_INFINITY
            if verbose:
                print_comment(f"Setting time limit: {soft} -- {hard}")
                resource.setrlimit(resource.RLIMIT_CPU, (soft, hard))
            else:
                resource.setrlimit(resource.RLIMIT_CPU, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))
        else:
            warnings.warn("CPU time limits using `resource` are not supported on Windows. Skipping hard limit.")

def wall_time(p: psutil.Process):
    return time.time() - p.create_time()

def mib_as_bytes(mib: int) -> int:
    return mib * 1024 * 1024

def mb_as_bytes(mb: int) -> int:
    return mb * 1000 * 1000

def bytes_as_mb(bytes: int) -> int:
    return bytes // (1000 * 1000)

def bytes_as_gb(bytes: int) -> int:
    return bytes // (1000 * 1000 * 1000)

def bytes_as_mb_float(bytes: int) -> float:
    return bytes / (1000 * 1000)

def bytes_as_gb_float(bytes: int) -> float:
    return bytes / (1000 * 1000 * 1000)

# def current_memory_usage() -> int: # returns bytes
#     # not really sure which memory measurement to use: https://stackoverflow.com/questions/7880784/what-is-rss-and-vsz-in-linux-memory-management/21049737#21049737
#     return psutil.Process(os.getpid()).memory_info().rss

def remaining_memory(limit:int) -> int: # bytes
    return limit # - current_memory_usage()

def get_subsolver(solver: str, model: cp.Model, subsolver: Optional[str] = None) -> Optional[str]:
    # Update subsolver
    if solver == "z3":
        if model.objective_ is not None:
            return "opt"
        else:
            return "sat"
    elif subsolver is None:
        if solver in SUPPORTED_SUBSOLVERS:
            return SUPPORTED_SUBSOLVERS[solver][0]
    return subsolver


# ---------------------------------------------------------------------------- #
#                            XCSP3 Output formatting                           #
# ---------------------------------------------------------------------------- #

def status_line_start() -> str:
    return 's' + chr(32)

def value_line_start() -> str:
    return 'v' + chr(32)

def objective_line_start() -> str:
    return 'o' + chr(32)

def comment_line_start() -> str:
    return 'c' + chr(32)

class ExitStatus(Enum):
    unsupported:str = "UNSUPPORTED" # instance contains an unsupported feature (e.g. a unsupported global constraint)
    sat:str = "SATISFIABLE" # CSP : found a solution | COP : found a solution but couldn't prove optimality
    optimal:str = "OPTIMUM" + chr(32) + "FOUND" # optimal COP solution found
    unsat:str = "UNSATISFIABLE" # instance is unsatisfiable
    unknown:str = "UNKNOWN" # any other case

def print_status(status: ExitStatus) -> None:
    print(status_line_start() + status.value, end="\n", flush=True)

def print_value(value: str) -> None:
    value = value[:-2].replace("\n", "\nv" + chr(32)) + value[-2:]
    print(value_line_start() + value, end="\n", flush=True)

def print_objective(objective: int) -> None:
    print(objective_line_start() + str(objective), end="\n", flush=True)

def print_comment(comment: str) -> None:
    print(comment_line_start() + comment.rstrip('\n'), end="\r\n", flush=True)

def solution_xml(model, useless_style="*", boolean_style="int"):
    """
        Formats a solution according to the XCSP3 specification.

        Arguments:
            model: CPMpy model for which to format its solution (should be solved first)
            useless_style: How to process unused decision variables (with value `None`). 
                           If "*", variable is included in reporting with value "*". 
                           If "drop", variable is excluded from reporting.
            boolean_style: Print style for boolean constants.
                           "int" results in 0/1, "bool" results in False/True.

        Returns:
            XML-formatted model solution according to XCSP3 specification.
    """

    # CSP
    if not model.has_objective():
        root = ET.Element("instantiation", type="solution")
    # COP
    else:
        root = ET.Element("instantiation", type="optimum", cost=str(int(model.objective_value())))

    # How useless variables should be handled
    #    (variables which have value `None` in the solution)
    variables = {var.name: var for var in model.user_vars if var.name[:2] not in [_IV_PREFIX, _BV_PREFIX, "B#"]} # dirty workaround for all missed aux vars in user vars
    if useless_style == "*":
        variables = {k:(v.value() if v.value() is not None else "*") for k,v in variables.items()}
    elif useless_style == "drop":
        variables = {k:v.value() for k,v in variables.items() if v.value() is not None}

    # Convert booleans
    if boolean_style == "bool":
        pass
    elif boolean_style == "int":
        variables = {k:(v if (not isinstance(v, bool)) else (1 if v else 0)) for k,v in variables.items()}

    # Build XCSP3 XML tree
    ET.SubElement(root, "list").text=" " + " ".join([str(v) for v in variables.keys()]) + " "
    ET.SubElement(root, "values").text=" " + " ".join([str(v) for v in variables.values()]) + " "
    tree = ET.ElementTree(root)
    ET.indent(tree, space="  ", level=0)
    res = ET.tostring(root).decode("utf-8")

    return str(res)


# ---------------------------------------------------------------------------- #
#                          CLI argument type checkers                          #
# ---------------------------------------------------------------------------- #

def dir_path(path):
    if os.path.isfile(path):
        return Path(path)
    else:
        raise argparse.ArgumentTypeError(f"readable_dir:{path} is not a valid path")

def is_supported_solver(solver:Optional[str]):
    if (solver is not None) and (solver not in SUPPORTED_SOLVERS):
        return False    
    else:
        return True
    
def is_supported_subsolver(solver, subsolver:Optional[str]):
    if (subsolver is not None) and (subsolver not in SUPPORTED_SUBSOLVERS.get(solver, [None])):
        return False
    else:
        return True

def supported_solver(solver:Optional[str]):
    if not is_supported_solver(solver):
        argparse.ArgumentTypeError(f"solver:{solver} is not a supported solver. Options are: {str(SUPPORTED_SOLVERS)}")
    else:
        return solver
    

# ---------------------------------------------------------------------------- #
#                         Executable & Solver arguments                        #
# ---------------------------------------------------------------------------- #

def ortools_arguments(model: cp.Model,
                      cores: Optional[int] = None,
                      seed: Optional[int] = None,
                      intermediate: bool = False,
                      **kwargs):
    # https://github.com/google/or-tools/blob/stable/ortools/sat/sat_parameters.proto
    res = dict()

    # https://github.com/google/or-tools/blob/1c5daab55dd84bca7149236e4b4fa009e5fd95ca/ortools/flatzinc/cp_model_fz_solver.cc#L1688
    res |= {
        "interleave_search": True,
        "use_rins_lns": False,
    }
    if not model.has_objective():
        res |= { "num_violation_ls": 1 }

    if cores is not None:
        res |= { "num_search_workers": cores }
    if seed is not None: 
        res |= { "random_seed": seed }

    if intermediate and model.has_objective():
        # Define custom ORT solution callback, then register it
        from ortools.sat.python import cp_model as ort
        class OrtSolutionCallback(ort.CpSolverSolutionCallback):
            """
                For intermediate objective printing.
            """

            def __init__(self):
                super().__init__()
                self.__start_time = time.time()
                self.__solution_count = 1

            def on_solution_callback(self):
                """Called on each new solution."""
                
                current_time = time.time()
                obj = int(self.ObjectiveValue())
                print_comment('Solution %i, time = %0.2fs' % 
                            (self.__solution_count, current_time - self.__start_time))
                print_objective(obj)
                self.__solution_count += 1
            

            def solution_count(self):
                """Returns the number of solutions found."""
                return self.__solution_count
            
        # Register the callback
        res |= { "solution_callback": OrtSolutionCallback() }

    def internal_options(solver: CPM_ortools):
        # https://github.com/google/or-tools/blob/1c5daab55dd84bca7149236e4b4fa009e5fd95ca/ortools/flatzinc/cp_model_fz_solver.cc#L1688
        solver.ort_solver.parameters.subsolvers.extend(["default_lp", "max_lp", "quick_restart"])
        if not model.has_objective():
            solver.ort_solver.parameters.subsolvers.append("core_or_no_lp")
        if len(solver.ort_model.proto.search_strategy) != 0:
            solver.ort_solver.parameters.subsolvers.append("fixed")

    return res, internal_options

def exact_arguments(seed: Optional[int] = None, **kwargs):
    # Documentation: https://gitlab.com/JoD/exact/-/blob/main/src/Options.hpp?ref_type=heads
    res = dict()
    if seed is not None: 
        res |= { "seed": seed }

    return res, None

def choco_arguments(): 
    # Documentation: https://github.com/chocoteam/pychoco/blob/master/pychoco/solver.py
    return {}, None

def z3_arguments(model: cp.Model,
                 cores: int = 1,
                 seed: Optional[int] = None,
                 mem_limit: Optional[int] = None,
                 **kwargs):
    # Documentation: https://microsoft.github.io/z3guide/programming/Parameters/
    # -> is outdated, just let it crash and z3 will report the available options

    res = dict()
    
    if model.has_objective():
        # Opt does not seem to support setting random seed or max memory
        pass
    else:
        # Sat parameters
        if cores is not None:
            res |= { "threads": cores }  # TODO what with hyperthreadding, when more threads than cores
        if seed is not None: 
            res |= { "random_seed": seed }
        if mem_limit is not None:
            res |= { "max_memory": bytes_as_mb(mem_limit) }

    return res, None

def minizinc_arguments(solver: str,
                       cores: Optional[int] = None,
                       seed: Optional[int] = None,
                       **kwargs):
    # Documentation: https://minizinc-python.readthedocs.io/en/latest/api.html#minizinc.instance.Instance.solve
    res = dict()
    if cores is not None:
        res |= { "processes": cores }
    if seed is not None: 
        res |= { "random_seed": seed }

    #if solver.endswith("gecode"):
        # Documentation: https://www.minizinc.org/doc-2.4.3/en/lib-gecode.html
    #elif solver.endswith("chuffed"):
        # Documentation: 
        # - https://www.minizinc.org/doc-2.5.5/en/lib-chuffed.html
        # - https://github.com/chuffed/chuffed/blob/develop/chuffed/core/options.h
    
    return res, None

def gurobi_arguments(model: cp.Model,
                     cores: Optional[int] = None,
                     seed: Optional[int] = None,
                     mem_limit: Optional[int] = None,
                     intermediate: bool = False,
                     **kwargs):
    # Documentation: https://www.gurobi.com/documentation/9.5/refman/parameters.html#sec:Parameters
    res = dict()
    if cores is not None:
        res |= { "Threads": cores }
    if seed is not None:
        res |= { "Seed": seed }
    if mem_limit is not None:
        res |= { "MemLimit": bytes_as_gb(remaining_memory(mem_limit)) }

    if intermediate and model.has_objective():

        class GurobiSolutionCallback:
            def __init__(self, model:cp.Model):
                self.__start_time = time.time()
                self.__solution_count = 0
                self.model = model

            def callback(self, *args, **kwargs):
                current_time = time.time()
                model, state = args

                # Callback codes: https://www.gurobi.com/documentation/current/refman/cb_codes.html#sec:CallbackCodes
                
                from gurobipy import GRB
                # if state == GRB.Callback.MESSAGE: # verbose logging
                #     print_comment("log message: " + str(model.cbGet(GRB.Callback.MSG_STRING)))
                if state == GRB.Callback.MIP: # callback from the MIP solver
                    if model.cbGet(GRB.Callback.MIP_SOLCNT) > self.__solution_count: # do we have a new solution?

                        obj = int(model.cbGet(GRB.Callback.MIP_OBJBST))
                        print_comment('Solution %i, time = %0.2fs' % 
                                    (self.__solution_count, current_time - self.__start_time))
                        print_objective(obj)
                        self.__solution_count = model.cbGet(GRB.Callback.MIP_SOLCNT)

        res |= { "solution_callback": GurobiSolutionCallback(model).callback }

    return res, None

def cpo_arguments(model: cp.Model,
                  cores: Optional[int] = None,
                  seed: Optional[int] = None,
                  intermediate: bool = False,
                  **kwargs):
    # Documentation: https://ibmdecisionoptimization.github.io/docplex-doc/cp/docplex.cp.parameters.py.html#docplex.cp.parameters.CpoParameters
    res = dict()
    if cores is not None:
        res |= { "Workers": cores }
    if seed is not None:
        res |= { "RandomSeed": seed }

    if intermediate and model.has_objective():
        from docplex.cp.solver.solver_listener import CpoSolverListener

        class CpoSolutionCallback(CpoSolverListener):

            def __init__(self):
                super().__init__()
                self.__start_time = time.time()
                self.__solution_count = 1

            def result_found(self, solver, sres):
                current_time = time.time()
                obj = sres.get_objective_value()
                if obj is not None:
                    print_comment('Solution %i, time = %0.2fs' % 
                                (self.__solution_count, current_time - self.__start_time))
                    print_objective(obj)
                    self.__solution_count += 1

            def solution_count(self):
                """Returns the number of solutions found."""
                return self.__solution_count

        # Register the callback
        res |= { "solution_callback": CpoSolutionCallback }

    return res, None


def solver_arguments(solver: str, 
                     model: cp.Model, 
                     seed: Optional[int] = None,
                     intermediate: bool = False,
                     cores: int = 1,
                     mem_limit: Optional[int] = None,
                     **kwargs):
    opt = model.has_objective()
    sat = not opt

    if solver == "ortools":
        return ortools_arguments(model, cores=cores, seed=seed, intermediate=intermediate, **kwargs)
    elif solver == "exact":
        return exact_arguments(seed=seed, **kwargs)
    elif solver == "choco":
        return choco_arguments()
    elif solver == "z3":
        return z3_arguments(model, cores=cores, seed=seed, mem_limit=mem_limit, **kwargs)
    elif solver.startswith("minizinc"):  # also can have a subsolver
        return minizinc_arguments(solver, cores=cores, seed=seed, **kwargs)
    elif solver == "gurobi":
        return gurobi_arguments(model, cores=cores, seed=seed, mem_limit=mem_limit, intermediate=intermediate, opt=opt, **kwargs)
    elif solver == "cpo":
        return cpo_arguments(model=model, cores=cores, seed=seed, intermediate=intermediate, **kwargs)
    else:
        print_comment(f"setting parameters of {solver} is not (yet) supported")
        return dict()

@contextmanager
def prepend_print():
    # Save the original stdout
    original_stdout = sys.stdout
    
    class PrependStream:
        def __init__(self, stream):
            self.stream = stream
        
        def write(self, message):
            # Prepend 'c' to each message before writing it
            if message.strip():  # Avoid prepending 'c' to empty messages (like newlines)
                self.stream.write('c ' + message)
            else:
                self.stream.write(message)
        
        def flush(self):
            self.stream.flush()
    
    # Override stdout with our custom stream
    sys.stdout = PrependStream(original_stdout)
    
    try:
        yield
    finally:
        # Restore the original stdout
        sys.stdout = original_stdout

 
# Run the instance; exceptions are caught and printed but also re-raised
def xcsp3_cpmpy(
        benchname: str,
        seed: Optional[int] = None,
        time_limit: Optional[int] = None,
        mem_limit: Optional[int] = None,  # MiB: 1024 * 1024 bytes
        cores: int = 1,
        solver: str = None,
        time_buffer: int = 0,
        intermediate: bool = False,
        verbose: bool = False,
        **kwargs,
):
    if not verbose:
        warnings.filterwarnings("ignore")

    try:

        # --------------------------- Global Configuration --------------------------- #

        # Get the current process
        p = psutil.Process()

        # pychoco currently does not support setting the mem_limit
        if solver == "choco" and mem_limit is not None:
            warnings.warn("'mem_limit' is currently not supported with choco, issues with GraalVM")
            mem_limit = None

        # Set random seed (if provided)
        if seed is not None:
            random.seed(seed)

        # Set memory limit (if provided)
        if mem_limit is not None:
            set_memory_limit(mem_limit, verbose=verbose)

        # Set time limit (if provided)
        if time_limit is not None:
            set_time_limit(int(time_limit - wall_time(p) + time.process_time()), verbose=verbose) # set remaining process time != wall time
   
        sys.argv = ["-nocompile"] # Stop pyxcsp3 from complaining on exit
        
        # ------------------------------ Parse instance ------------------------------ #

        time_parse = time.time()
        parser = _parse_xcsp3(benchname)
        time_parse = time.time() - time_parse
        if verbose: print_comment(f"took {time_parse:.4f} seconds to parse XCSP3 model [{benchname}]")

        if time_limit and time_limit < wall_time(p):
            raise TimeoutError("Time's up after parse")
                
        # ---------------- Convert XCSP3 to CPMpy model with callbacks --------------- #

        time_callback = time.time()
        model = _load_xcsp3(parser)
        time_callback = time.time() - time_callback
        if verbose: print_comment(f"took {time_callback:.4f} seconds to convert to CPMpy model")
        
        if time_limit and time_limit < wall_time(p):
            raise TimeoutError("Time's up after callback")
        
        # ------------ Replace solver supported constraints with natives ------------- #

        # Additional XCSP3-specific native constraints
        added_natives = {
            "ortools": {
                "no_overlap2d": globals.OrtNoOverlap2D,
                "subcircuit": globals.OrtNoOverlap2D,
                "subcircuitwithstart": lambda args: globals.OrtSubcircuitWithStart(args[:-1], args[-1]),
            },
            "choco": {
                "subcircuit": globals.ChocoSubcircuit,
            },
            "minizinc": {
                "subcircuit": globals.MinizincSubcircuit,
                "subcircuitwithstart": globals.MinizincSubcircuitWithStart,
            },
        }

        # Loop through all constraints and replace with native if supported
        if solver in added_natives:
            for i, constraint in enumerate(model.constraints):
                if constraint.name in added_natives[solver]:
                    model.constraints[i] = added_natives[solver][constraint.name](constraint.args)

        # ------------------------ Post CPMpy model to solver ------------------------ #

        solver_args, internal_options = solver_arguments(solver, model=model, seed=seed,
                                       intermediate=intermediate,
                                       cores=cores, mem_limit=mib_as_bytes(mem_limit) if mem_limit is not None else None,
                                       **kwargs)
        # time_limit is generic for all, done later

        # Post model to solver
        time_post = time.time()

        if solver == "exact": # Exact2 takes its options at creation time
            s = cp.SolverLookup.get(solver, model, **solver_args)
            solver_args = dict()  # no more solver args needed
        else:
            s = cp.SolverLookup.get(solver, model)
        time_post = time.time() - time_post
        if verbose: print_comment(f"took {time_post:.4f} seconds to post model to {solver}")

        if time_limit and time_limit < wall_time(p):
            raise TimeoutError("Time's up after post")

        # ------------------------------- Solve model ------------------------------- #
        
        if time_limit:
            # give solver only the remaining time
            time_limit = time_limit - wall_time(p) - time_buffer
            # disable signal-based time limit and let the solver handle it (solvers don't play well with difference between cpu and wall time)
            set_time_limit(None)
            
            if verbose: print_comment(f"{time_limit}s left to solve")
        
        time_solve = time.time()
        try:
            if internal_options is not None:
                internal_options(s) # Set more internal solver options (need access to native solver object)
            is_sat = s.solve(time_limit=time_limit, **solver_args)
        except RuntimeError as e:
            if "Program interrupted by user." in str(e): # Special handling for Exact
                raise TimeoutError("Exact interrupted due to timeout")
            else:
                raise e

        time_solve = time.time() - time_solve
        if verbose: print_comment(f"took {time_solve:.4f} seconds to solve")

        # ------------------------------- Print result ------------------------------- #

        if s.status().exitstatus == CPMStatus.OPTIMAL:
            print_value(solution_xml(s))
            print_status(ExitStatus.optimal)
        elif s.status().exitstatus == CPMStatus.FEASIBLE:
            print_value(solution_xml(s))
            print_status(ExitStatus.sat)
        elif s.status().exitstatus == CPMStatus.UNSATISFIABLE:
            print_status(ExitStatus.unsat)
        else:
            print_comment("Solver did not find any solution within the time/memory limit")
            print_status(ExitStatus.unknown)

        # ------------------------------------- - ------------------------------------ #

        
    except MemoryError as e:
        print_comment(f"MemoryError raised. Reached limit of {mem_limit} MiB")
        print_status(ExitStatus.unknown)
        raise e
    except ParseError as e:
        if "out of memory" in e.msg:
            print_comment(f"MemoryError raised by parser. Reached limit of {mem_limit} MiB")
            print_status(ExitStatus.unknown)
        else:
            print_comment(f"An {type(e)} got raised: {e}")
            print_status(ExitStatus.unknown)
        raise e
    except NotImplementedError as e:
        print_comment(str(e))
        print_status(ExitStatus.unsupported)
        raise e
    except Exception as e:
        print_comment(f"An {type(e)} got raised: {e}")
        import traceback
        print_comment("Stack trace:")
        for line in traceback.format_exc().split('\n'):
            if line.strip():
                print_comment(line)
        print_status(ExitStatus.unknown)
        raise e


if __name__ == "__main__":
    

    # ------------------------------ Argument parsing ------------------------------ #

    parser = argparse.ArgumentParser("CPMpy XCSP3 executable")
    
    ## XCSP3 required arguments:
    # BENCHNAME: Name of the XCSP3 XML file with path and extension
    parser.add_argument("benchname", type=dir_path, help="XCSP3 XML file to parse, with full path and extension.") 
    # RANDOMSEED: Seed between 0 and 4294967295
    parser.add_argument("-s", "--seed", required=False, type=int, default=None, help="Random seed (integer between 0 and 4294967295).")
    # TIMELIMIT: Total CPU time in seconds (before it gets killed)
    parser.add_argument("-l", "--time-limit", required=False, type=int, default=None, help="Time limit in seconds.") # TIMELIMIT
    # MEMLIMIT: Total amount of memory in MiB (mebibyte = 1024 * 1024 bytes)
    parser.add_argument("-m", "--mem-limit", required=False, type=int, default=None, help="Memory limit in MiB (1 MiB = 1024 x 1024 bytes). Gets passed on to the solver (if supported) and is also enforced on the Python/CPMpy side. Measured as Virtual Memory Size")
    # TMPDIR: Only location where temporary read/write is allowed
    parser.add_argument("-t","--tmpdir", required=False, type=dir_path, help="Directory for temporary read/write operations.")
    # NBCORE: Number of processing units (can by any of the following: a processor / a processor core / logical processor (hyper-threading))
    parser.add_argument("-c", "--cores", required=False, type=int, default=None, help="Number of processing units to use (e.g., cores or logical processors).")
    # DIR: not needed, e.g. we just import files

    ## CPMpy optional arguments:
    # The underlying solver which should be used (can also be "solver:subsolver")
    parser.add_argument("--solver", required=False, type=str, default="ortools", help="Underlying CPMpy solver to use (can be 'solver:subsolver').")
    # How much time before SIGTERM should we halt solver (for the final post-processing steps and solution printing)
    parser.add_argument("--time-buffer", required=False, type=int, default=0, help="Time buffer (in seconds) to reserve before SIGTERM for cleanup/postprocessing.") # buffer to give subsolver time to return and do some post-processing
    # If intermediate solutions should be printed (if the solver supports it)
    parser.add_argument("--intermediate", action=argparse.BooleanOptionalAction, help="Whether to print intermediate solutions (if supported by the solver).")

    ## Executable optional arguments
    parser.add_argument("--verbose", action=argparse.BooleanOptionalAction, help="Enable verbose output for debugging or detailed logs.")
    

    # Process cli arguments 
    args = parser.parse_args()
    if args.verbose:
        print_comment(f"Arguments: {args}")

    try:
        # Configure signal handles
        init_signal_handlers()

        if str(args.benchname).endswith(".lzma"):
            # Decompress the XZ file
            with lzma.open(args.benchname, 'rt', encoding='utf-8') as f:
                xml_file = StringIO(f.read()) # read to memory-mapped file
                args.benchname = xml_file

        xcsp3_cpmpy(**vars(args))
    except Exception as e:
        print_comment(f"{type(e).__name__} -- {e}")


"""
    Decompose global constraints not supported by the solver.
"""

import copy
import warnings  # for deprecation warning

from .normalize import toplevel_list
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.globalfunctions import GlobalFunction
from ..expressions.core import Expression, Comparison, Operator
from ..expressions.variables import intvar, cpm_array, NDVarArray
from ..expressions.utils import is_any_list, eval_comparison
from ..expressions.python_builtins import all
from .flatten_model import flatten_constraint, normalized_numexpr


def decompose_in_tree(lst_of_expr, supported=set(), supported_reified=set(), _toplevel=None, nested=False, csemap=None):
    """
        Decomposes any global constraint not supported by the solver.
        Accepts a list of CPMpy expressions as input and returns a list of CPMpy expressions.
            
        :param supported: a set of supported global constraints or global functions
        :param supported_reified: a set of supported reified global constraints (globals with Boolean return type only)
        :param _toplevel: a list of constraints that should be added toplevel, carried as pass by reference to recursive calls
        :param nested: if True, new constraints will have been added to the `_toplevel` list too
        
        Special care taken for unsupported global constraints in reified contexts and for numeric global constraints
        in a comparison.

        Supported numerical global functions remain in the expression tree as is. They can be rewritten using
        :func:`cpmpy.transformations.reification.reify_rewrite`
        The following ``bv -> NumExpr <comp> Var/Const`` can be rewritten as  ``[bv -> IV0 <comp> Var/Const, NumExpr == IV0]``.
        So even if numerical constraints are not supported in reified context, we can rewrite them to non-reified versions if they are total.
    """
    if _toplevel is None:
        _toplevel = []

    # swap the arguments of a comparison while maintaining its semantics
    flipmap = {"==": "==", "!=": "!=", "<": ">", "<=": ">=", ">": "<", ">=": "<="}

    newlist = []  # decomposed constraints will go here
    for expr in lst_of_expr:

        if is_any_list(expr):
            assert nested is True, "Cannot have nested lists without passing trough an expression, make sure to run " \
                                   "cpmpy.transformations.normalize.toplevel_list first. "
            if isinstance(expr, NDVarArray):  # NDVarArray is also an expression,
                                              # so we can call has_subexpr on it for a possible early-exit
                if expr.has_subexpr():
                    newexpr = decompose_in_tree(expr, supported, supported_reified, _toplevel, nested=True, csemap=csemap)
                    newlist.append(cpm_array(newexpr))
                else:
                    newlist.append(expr)
            else: # a normal list-like (list, tuple, np.ndarray), must be called recursively and check all elements
                newexpr = decompose_in_tree(expr, supported, supported_reified, _toplevel, nested=True, csemap=csemap)
                newlist.append(newexpr)
            continue

        elif isinstance(expr, Operator):

            if not expr.has_subexpr(): # Only recurse if there are nested expressions
                newlist.append(expr)
                continue

            if any(isinstance(a,GlobalFunction) for a in expr.args):
                expr, base_con = normalized_numexpr(expr, csemap=csemap)
                _toplevel.extend(base_con)  # should be added toplevel
            # recurse into arguments, recreate through constructor (we know it stores no other state)
            args = decompose_in_tree(expr.args, supported, supported_reified, _toplevel, nested=True, csemap=csemap)
            newlist.append(Operator(expr.name, args))

        elif isinstance(expr, GlobalConstraint) or isinstance(expr, GlobalFunction):
            # Can't early-exit here, need to check if constraint in itself is even supported
            if nested and expr.is_bool():
                # special case: reified (Boolean) global
                is_supported = (expr.name in supported_reified)
            else:
                is_supported = (expr.name in supported)

            if is_supported:
                # If no nested expressions, don't recurse the arguments
                if not expr.has_subexpr():
                    newlist.append(expr)
                    continue
                # Recursively decompose the subexpression arguments
                else:
                    expr = copy.copy(expr)
                    expr.update_args(decompose_in_tree(expr.args, supported, supported_reified, _toplevel, nested=True, csemap=csemap))
                    newlist.append(expr)

            else:
                if expr.is_bool():
                    assert isinstance(expr, GlobalConstraint)
                    # boolean global constraints
                    dec = expr.decompose()
                    if not isinstance(dec, tuple):
                        warnings.warn(f"Decomposing an old-style global ({expr}) that does not return a tuple, which is "
                                      "deprecated. Support for old-style globals will be removed in stable version",
                                      DeprecationWarning)
                        dec = (dec, [])
                    decomposed, define = dec

                    _toplevel.extend(define)  # definitions should be added toplevel
                    # the `decomposed` expression might contain other global constraints, check it
                    decomposed = decompose_in_tree(decomposed, supported, supported_reified, _toplevel, nested=nested, csemap=csemap)
                    newlist.append(all(decomposed))

                else:
                    # global function, replace by a fresh variable and decompose the equality to this
                    assert isinstance(expr, GlobalFunction)
                    lb,ub = expr.get_bounds()
                    
                    if csemap is not None and expr in csemap:
                        aux = csemap[expr]
                    else:
                        aux = intvar(lb, ub)
                        if csemap is not None:
                            csemap[expr] = aux

                    # NOTE Do we need to decompose here (the expr's args)? As done in the Comparison's section?

                    dec = expr.decompose_comparison("==", aux)
                    if not isinstance(dec, tuple):
                        warnings.warn(f"Decomposing an old-style global ({expr}) that does not return a tuple, which is "
                                      "deprecated. Support for old-style globals will be removed in stable version",
                                      DeprecationWarning)
                        dec = (dec, [])
                    auxdef, otherdef = dec

                    _toplevel.extend(auxdef + otherdef)  # all definitions should be added toplevel
                    newlist.append(aux)  # replace original expression by aux

        elif isinstance(expr, Comparison):
            if not expr.has_subexpr(): # Only recurse if there are nested expressions
                newlist.append(expr)
                continue

            # if one of the two children is a (numeric) global constraint, we can decompose the comparison directly
            # otherwise e.g., min(x,y,z) == a would become `min(x,y,z).decompose_comparison('==',aux) + [aux == a]`
            lhs, rhs = expr.args
            exprname = expr.name  # can change when rhs needs decomp

            decomp_lhs = isinstance(lhs, GlobalFunction) and not lhs.name in supported
            decomp_rhs = isinstance(rhs, GlobalFunction) and not rhs.name in supported

            if not decomp_lhs:
                if not decomp_rhs:
                    # nothing special, create a fresh version and recurse into arguments
                    expr = copy.copy(expr)
                    expr.update_args(decompose_in_tree(expr.args, supported, supported_reified, _toplevel, nested=True, csemap=csemap))
                    newlist.append(expr)

                else:
                    # only rhs needs decomposition, so flip comparison to make lhs needing the decomposition
                    exprname = flipmap[expr.name]
                    lhs, rhs = rhs, lhs
                    decomp_lhs, decomp_rhs = True, False  # continue into next 'if'

            if decomp_lhs:
                if lhs.has_subexpr():
                    # recurse into lhs args and decompose nested subexpressions
                    lhs = copy.copy(lhs)
                    lhs.update_args(decompose_in_tree(lhs.args, supported, supported_reified, _toplevel, nested=True, csemap=csemap))

                # decompose comparison of lhs and rhs
                dec = lhs.decompose_comparison(exprname, rhs)
                if not isinstance(dec, tuple):
                    warnings.warn(f"Decomposing an old-style global ({lhs}) that does not return a tuple, which is "
                                  f"deprecated. Support for old-style globals will be removed in stable version",
                                  DeprecationWarning)
                    dec = (dec, [])
                decomposed, define = dec

                _toplevel.extend(define)  # definitions should be added toplevel
                # the `decomposed` expression (and rhs) might contain other global constraints, check it
                decomposed = decompose_in_tree(decomposed, supported, supported_reified, _toplevel, nested=True, csemap=csemap)
                newlist.append(all(decomposed))

        else:  # constants, variables, direct constraints
            newlist.append(expr)

    if nested:
        return newlist

    if len(_toplevel) == 0:
        return toplevel_list(newlist)
    else:
        # we are toplevel and some new constraints are introduced, decompose new constraints!
        return toplevel_list(newlist) + decompose_in_tree(toplevel_list(_toplevel), supported, supported_reified, nested=False, csemap=csemap)


# DEPRECATED!
# old way of doing decompositions post-flatten
# will be removed in any future version!
def decompose_global(lst_of_expr, supported=set(), supported_reif=set()):
    """
    .. deprecated:: 0.9.16
          Please use :func:`decompose_in_tree()` instead.
    """
    warnings.warn("Deprecated, use `decompose_in_tree()` instead, will be removed in stable version", DeprecationWarning)
    """
        DEPRECATED!!! USE `decompose_in_tree()` instead!
        Decomposes any global constraint not supported by the solver
        Accepts a list of flat constraints as input
        Returns a list of flat constraints

        - supported: a set of supported global constraints or global functions
        - supported_reified: a set of supported global constraints within a reification

        Unsupported global constraints are decomposed into equivalent simpler constraints.
        Numerical global constraints are left reified even if not part of the `supported_reified` set
            as they can be rewritten using `cpmpy.transformations.reification.reify_rewrite`...
                ... reified partial functions are decomposed when unsupported by `supported_reified`
                The following `bv -> NumExpr <comp> Var/Const` can always be rewritten as ...
                                [bv -> IV0 <comp> Var/Const, NumExpr == IV0].
                So even if numerical constraints are not supported in reified context, we can rewrite them to non-reified versions.
                TODO: decide if we want to do the rewrite of unsupported globals here or leave it to `reify_rewrite`.
                    Currently, we left it for `reify_rewrite`



    """
    def _is_supported(cpm_expr, reified):
        if isinstance(cpm_expr, GlobalConstraint):
            if reified and cpm_expr.name not in supported_reif:
                return False
            if not reified and cpm_expr.name not in supported:
                return False
        if isinstance(cpm_expr, Comparison) and isinstance(cpm_expr.args[0], GlobalFunction):
            if not cpm_expr.args[0].name in supported:
                # reified numerical global constraints can be rewritten to non-reified versions
                #  so only have to check for 'supported' set
                return False
            if reified and not cpm_expr.args[0].is_total() and cpm_expr.args[0].name not in supported_reif:
                # edge case for partial functions as they cannot be rewritten to non-reified versions
                #  have to decompose to total functions
                #  ASSUMPTION: the decomposition for partial global functions is total! (for now only Element)
                return False

        return True

    assert supported_reif <= supported, "`supported_reif` set is assumed to be a subset of the `supported` set"
    if not is_any_list(lst_of_expr):
        lst_of_expr= [lst_of_expr]

    newlist = []
    for cpm_expr in lst_of_expr:
        assert isinstance(cpm_expr, Expression), f"Expected CPMpy expression but got {cpm_expr}, run 'cpmpy.transformations.normalize.toplevel_list' first"
        decomp_idx = None

        if hasattr(cpm_expr, "decompose") and not _is_supported(cpm_expr, reified=False):
            cpm_expr = cpm_expr.decompose() # base boolean global constraints
        elif isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            if cpm_expr.name == "==" and not _is_supported(lhs, reified=True): # rhs will always be const/var
                decomp_idx = 0 # this means it is an unsupported reified boolean global

            if not _is_supported(cpm_expr, reified=False):
                cpm_expr = do_decompose(cpm_expr) # unsupported numerical constraint in comparison
                decomp_idx = None

        elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            lhs, rhs = cpm_expr.args # BV -> Expr or Expr -> BV as flat normal form is required
            if not _is_supported(rhs, reified=True):
                decomp_idx = 1 # reified (numerical) global constraint, probably most frequent case
            elif not _is_supported(lhs, reified=True):
                decomp_idx = 0

        if decomp_idx is not None:
            cpm_expr = copy.deepcopy(cpm_expr) # need deepcopy as we are changing args of list inplace
            cpm_expr.args[decomp_idx] = all(do_decompose(cpm_expr.args[decomp_idx]))
            cpm_expr = [cpm_expr]

        if isinstance(cpm_expr, list): # some decomposition happened, have to run again as decomp can contain new global
            flat = flatten_constraint(cpm_expr) # most of the time will already be flat... do check here?
            newlist.extend(decompose_global(flat, supported=supported))
        else:
            # default
            newlist.append(cpm_expr)

    return newlist

def do_decompose(cpm_expr):
    """
    .. deprecated:: 0.9.13
          Please use :func:`decompose_in_tree()` instead.
    """
    warnings.warn("Deprecated, never meant to be used outside this transformation; will be removed in stable version",
                  DeprecationWarning)
    """
        DEPRECATED
        Helper function for decomposing global constraints
        - cpm_expr: Global constraint or comparison containing a global constraint on lhs
    """
    if isinstance(cpm_expr, Comparison):
        lhs, rhs = cpm_expr.args
        if lhs.is_bool():
            return [eval_comparison(cpm_expr.name, all(lhs.decompose()), rhs)]
        else:
            dec = lhs.decompose_comparison(cpm_expr.name, rhs)
            # new style is a tuple of size 2, old style is a single list (which captures less cases)
            if isinstance(dec, tuple):
                dec = dec[0]+dec[1]
            return dec

    dec = cpm_expr.decompose()
    # new style is a tuple of size 2, old style is a single list (which captures less cases)
    if isinstance(dec, tuple):
        dec = dec[0]+dec[1]
    return dec



# /usr/bin/env python
"""
Interface to the Pindakaas solver's Python API.

Pindakaas is an open-source Rust library for encoding propositional and pseudo-Boolean constraints to SAT, with support for incremental solving and assumptions.

Always use :func:`cp.SolverLookup.get("pindakaas") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

============
Installation
============

Requires that the 'pindakaas' optional dependency is installed:

.. code-block:: console

    $ pip install pindakaas

Detailed installation instructions available at:

- https://pypi.org/project/pindakaas/
- https://github.com/pindakaashq/pindakaas

The rest of this documentation is for advanced users.

===============
List of classes
===============

.. autosummary::
    :nosignatures:

    CPM_pindakaas

==============
Module details
==============
"""

import time
import pkg_resources
from datetime import timedelta
from typing import Optional

from ..exceptions import NotSupportedError
from ..expressions.utils import eval_comparison
from ..expressions.core import BoolVal, Comparison
from ..expressions.variables import NegBoolView, _BoolVarImpl, _IntVarImpl
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.flatten_model import flatten_constraint
from ..transformations.get_variables import get_variables
from ..transformations.int2bool import _decide_encoding, _encode_int_var, int2bool
from ..transformations.linearize import linearize_constraint
from ..transformations.normalize import simplify_boolean, toplevel_list
from ..transformations.reification import only_bv_reifies, only_implies
from ..transformations.safening import no_partial_functions
from .solver_interface import ExitStatus, SolverInterface


class CPM_pindakaas(SolverInterface):
    """
    Interface to Pindakaas' Python API.

    Creates the following attributes (see parent constructor for more):

    - ``pdk_solver``: The Pindakaas solver back-end which encodes and solves models through the SAT sub-solver
    - ``ivarmap``: a mapping from integer variables to their encoding for `int2bool`
    - ``encoding``: the encoding used for `int2bool`, choose from ("auto", "direct", "order", or "binary"). Set to "auto" but can be changed in the solver object.
    - `unsatisfiable`: if a constraint is found to be unsatisfiable during the encoding phase, this flag is set to `True` to prevent further encoding efforts
    - ``core``: if the problem is unsatisfiable, the unsatisfiable core, else `None`


    Documentation of the solver's own Python API:

    - https://pypi.org/project/pindakaas/
    - https://github.com/pindakaashq/pindakaas

    """

    @staticmethod
    def supported():
        try:
            import pindakaas

            return True
        except ModuleNotFoundError:
            return False

    @staticmethod
    def version() -> Optional[str]:
        """Return the installed version of the solver's Python API."""
        try:
            return pkg_resources.get_distribution("pindakaas").version
        except pkg_resources.DistributionNotFound:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Initialize Pindakaas interface.

        """
        name = "pindakaas"
        if not self.supported():
            raise ImportError(
                f"CPM_{name}: Install the Pindakaas python library `pindakaas` (e.g. `pip install pindakaas`) package to use this solver interface"
            )
        if cpm_model and cpm_model.objective_ is not None:
            raise NotSupportedError(f"CPM_{name}: only satisfaction, does not support an objective function")

        import pindakaas as pdk

        assert subsolver is None, "Pindakaas does not support any subsolvers for the moment"
        self.ivarmap = dict()  # for the integer to boolean encoders
        self.encoding = "auto"
        self.pdk_solver = pdk.solver.CaDiCaL()
        self.unsatisfiable = False  # `pindakaas` might determine unsat before solving
        self.core = None  # latest UNSAT core
        super().__init__(name=name, cpm_model=cpm_model)

    @property
    def native_model(self):
        return self.pdk_solver

    def solve(self, time_limit=None, assumptions=None):
        """
        Solve the encoded CPMpy model given optional time limit and assumptions, returning whether a solution was found.

        :param time_limit: optional, time limit in seconds
        :param assumptions: optional, a list of assumptions (Boolean variables which should hold for this solve call)
        """
        if self.unsatisfiable:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            return self._solve_return(self.cpm_status)

        if time_limit is not None and time_limit <= 0:
            raise ValueError("Time limit must be positive")

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # the user vars are only the Booleans (e.g. to ensure solveAll behaves consistently)
        user_vars = set()
        for x in self.user_vars:
            if isinstance(x, _BoolVarImpl):
                user_vars.add(x)
            else:
                # extends set with encoding variables of `x`
                user_vars.update(self.ivarmap[x.name].vars())

        self.user_vars = user_vars

        if time_limit is not None:
            time_limit = timedelta(seconds=time_limit)
        solver_assumptions = None if assumptions is None else self.solver_vars(assumptions)

        t = time.time()
        with self.pdk_solver.solve(
            time_limit=time_limit,
            assumptions=solver_assumptions,
        ) as result:
            self.cpm_status.runtime = time.time() - t

            # translate pindakaas result status to cpmpy status
            import pindakaas as pdk

            if result.status == pdk.solver.Status.SATISFIED:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif result.status == pdk.solver.Status.UNSATISFIABLE:
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            elif result.status == pdk.solver.Status.UNKNOWN:
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                raise NotImplementedError(f"Pindakaas returned an unkown type of result status: {result}")

            # True/False depending on self.cpm_status
            has_sol = self._solve_return(self.cpm_status)

            # translate solution values (of user specified variables only)
            if has_sol:
                # fill in variable values
                for cpm_var in self.user_vars:
                    # essentially `.solver_var`, but failing if new vars are added
                    if isinstance(cpm_var, NegBoolView):
                        lit = ~self._varmap[cpm_var._bv.name]
                    elif isinstance(cpm_var, _BoolVarImpl):
                        lit = self._varmap[cpm_var.name]
                    else:
                        raise ValueError(
                            f"Integer variables should have been encoded using `int2bool` transformation, but {cpm_var} is integer, please report on GitHub"
                        )
                    value = result.value(lit)
                    assert value is not None, (
                        "All user variables should have been assigned, but {cpm_var} (literal {lit}) was not."
                    )
                    cpm_var._value = value
                self.core = None
                # Now assign the user integer variables using their encodings
                # `ivarmap` also contains auxiliary variable, but they will be assigned 'None' as their encoding variables are assigned `None`
                for enc in self.ivarmap.values():
                    enc._x._value = enc.decode()

            else:  # clear values of variables
                for cpm_var in self.user_vars:
                    cpm_var._value = None
                # we have to save the unsat core here, as the result object does not live beyond this solve call
                if assumptions is not None:
                    self.core = [x for x, s_x in zip(assumptions, solver_assumptions) if result.failed(s_x)]

        return has_sol

    def solver_var(self, cpm_var):
        if isinstance(cpm_var, NegBoolView):  # negative literal
            # get inner variable and return its negated solver var
            return ~self.solver_var(cpm_var._bv)
        elif isinstance(cpm_var, _BoolVarImpl):  # positive literal
            # insert if new
            if cpm_var.name not in self._varmap:
                self._varmap[cpm_var.name] = self.pdk_solver.new_var()
            return self._varmap[cpm_var.name]
        elif isinstance(cpm_var, _IntVarImpl):  # intvar
            if cpm_var.name not in self.ivarmap:
                enc, cons = _encode_int_var(
                    self.ivarmap, cpm_var, _decide_encoding(cpm_var, None, encoding=self.encoding)
                )
                self += cons
            else:
                enc = self.ivarmap[cpm_var.name]
            return self.solver_vars(enc.vars())
        else:
            raise TypeError

    def transform(self, cpm_expr):
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons)
        cpm_cons = decompose_in_tree(cpm_cons, csemap=self._csemap)
        cpm_cons = simplify_boolean(cpm_cons)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)
        cpm_cons = linearize_constraint(
            cpm_cons, supported=frozenset({"sum", "wsum", "and", "or"}), csemap=self._csemap
        )
        cpm_cons = int2bool(cpm_cons, self.ivarmap, encoding=self.encoding)
        return cpm_cons

    def add(self, cpm_expr_orig):
        import pindakaas as pdk

        if self.unsatisfiable:
            return self

        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        # transform and post the constraints
        try:
            for cpm_expr in self.transform(cpm_expr_orig):
                self._post_constraint(cpm_expr)
        except pdk.Unsatisfiable:
            self.unsatisfiable = True

        return self

    __add__ = add  # avoid redirect in superclass

    def _post_constraint(self, cpm_expr, conditions=[]):
        """Add a single, *transformed* constraint, implied by conditions."""
        if isinstance(cpm_expr, BoolVal):
            # base case: Boolean value
            if cpm_expr.args[0] is False:
                self.pdk_solver.add_clause(conditions)

        elif isinstance(cpm_expr, _BoolVarImpl):  # (implied) literal
            self.pdk_solver.add_clause(conditions + [self.solver_var(cpm_expr)])

        elif cpm_expr.name == "or":  # (implied) clause
            self.pdk_solver.add_clause(conditions + self.solver_vars(cpm_expr.args))

        elif cpm_expr.name == "->":  # implication
            a0, a1 = cpm_expr.args
            self._post_constraint(a1, conditions=conditions + [~self.solver_var(a0)])

        elif isinstance(cpm_expr, Comparison):  # Bool linear
            assert cpm_expr.name in {"<=", ">=", "=="}, (
                f"Unsupported comparator {cpm_expr.name} for constraint should have been transformed: {cpm_expr}"
            )

            # lhs is a sum/wsum, right hand side a constant int
            lhs, rhs = cpm_expr.args
            if isinstance(lhs, _BoolVarImpl):
                literals = [lhs]
                coefficients = [1]
            elif lhs.name == "sum":
                literals = lhs.args
                coefficients = [1] * len(literals)
            elif lhs.name == "wsum":
                coefficients, literals = lhs.args
            else:
                raise ValueError(f"Trying to encode non (Boolean) linear constraint: {cpm_expr}")

            lhs = sum(c * l for c, l in zip(coefficients, self.solver_vars(literals)))

            self.pdk_solver.add_encoding(eval_comparison(cpm_expr.name, lhs, rhs), conditions=conditions)
        else:
            raise NotSupportedError(f"{self.name}: Unsupported constraint {cpm_expr}")

    def get_core(self):
        assert self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE and self.core is not None, (
            "get_core(): requires a previous solve call with assumption variables and an UNSATISFIABLE result"
        )
        return self.core



"""
    Solver interfaces have the same API as :class:`Model <cpmpy.model.Model>`.
    However some solvers are **incremental**, meaning that after solving a problem, you can add constraints or change
    the objective, and the next solve will reuse as much information from the previous solve as possible.
    Some clause learning solvers also support solving with **assumptions**, meaning you can solve the same problem with
    different assumption variables toggled on/off, and the solver will reuse information from the previous solves.

    See :ref:`supported-solvers` for the list of solvers and their capabilities.

    To benefit from incrementality, you have to instantiate the solver object and reuse it, rather than working on a Model object.
    Solvers must be instantiated throught the static :class:`cp.SolverLookup <cpmpy.solvers.utils.SolverLookup>` class:

    - :meth:`cp.SolverLookup.solvernames() <cpmpy.solvers.utils.SolverLookup.solvernames>` -- List all installed solvers (including subsolvers).
    - :meth:`cp.SolverLookup.get(solvername, model=None) <cpmpy.solvers.utils.SolverLookup.get>` -- Initialize a specific solver.

    For example creating a CPMpy solver object for OR-Tools:

    .. code-block:: python

        import cpmpy as cp
        s = cp.SolverLookup.get("ortools")
        # can now use solver object 's' over and over again 


    =========================
    List of solver submodules
    =========================
    .. autosummary::
        :nosignatures:

        ortools
        choco
        gcs
        minizinc
        cpo
        gurobi
        exact
        z3
        pysat
        pysdd
        pindakaas
        pumpkin
        cplex
        hexaly

    =========================
    List of helper submodules
    =========================
    .. autosummary::
        :nosignatures:

        solver_interface
        utils

    =================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        param_combinations
"""

from .utils import builtin_solvers, get_supported_solvers, param_combinations
from .ortools import CPM_ortools
from .pysat import CPM_pysat
from .minizinc import CPM_minizinc
from .gurobi import  CPM_gurobi
from .pysdd import CPM_pysdd
from .z3 import CPM_z3
from .exact import CPM_exact
from .choco import CPM_choco
from .gcs import CPM_gcs
from .cpo import CPM_cpo
from .pindakaas import CPM_pindakaas
from .pumpkin import CPM_pumpkin
from .cplex import CPM_cplex
from .hexaly import CPM_hexaly



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## __init__.py
##
"""
    Collection of tools for explanation techniques.

    =============
    List of tools
    =============

    .. autosummary::
        :nosignatures:

        mus
        mss
        mcs
        marco
        utils
"""

from .mus import *
from .mss import *
from .mcs import *
from .marco import *


"""Convert integer linear constraints to pseudo-boolean constraints."""

from typing import List
import itertools
import math
import cpmpy as cp
from abc import ABC, abstractmethod
from ..expressions.variables import _BoolVarImpl, _IntVarImpl, boolvar
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.core import Comparison, Operator, BoolVal
from ..expressions.core import Expression

UNKNOWN_COMPARATOR_ERROR = ValueError("Comparator is not known or should have been simplified by linearize.")
EMPTY_DOMAIN_ERROR = ValueError("Attempted to encode variable with empty domain (which is unsat)")


def int2bool(cpm_lst: List[Expression], ivarmap, encoding="auto"):
    """Convert integer linear constraints to pseudo-boolean constraints. Requires `linearize` transformation."""
    assert encoding in (
        "auto",
        "direct",
        "order",
        "binary",
    ), "Only auto, direct, order, and binary encoding are supported"

    cpm_out = []
    for expr in cpm_lst:
        constraints, domain_constraints = _encode_expr(ivarmap, expr, encoding)
        cpm_out += domain_constraints + constraints
    return cpm_out


def _encode_expr(ivarmap, expr, encoding):
    """Return encoded constraints and root-level constraints (e.g. domain constraints exactly-one, ..)."""
    constraints = []
    domain_constraints = []

    # skip all Boolean expressions
    if isinstance(expr, (BoolVal, _BoolVarImpl, DirectConstraint)) or expr.name == "or":
        return [expr], []
    elif expr.name == "->":
        # Encode implication recursively
        p, consequent = expr.args
        constraints, domain_constraints = _encode_expr(ivarmap, consequent, encoding)
        return (
            [p.implies(constraint) for constraint in constraints],
            domain_constraints,
        )
    elif isinstance(expr, Comparison):
        lhs, rhs = expr.args
        # Encode linears with single left-hand term using more efficient `_encode_comparison`
        if type(lhs) is _BoolVarImpl:
            return [expr], []
        elif type(lhs) is _IntVarImpl:
            return _encode_comparison(ivarmap, lhs, expr.name, rhs, encoding)
        elif lhs.name == "sum":
            if len(lhs.args) == 1:
                return _encode_expr(
                    ivarmap, Comparison(expr.name, lhs.args[0], rhs), encoding
                )  # even though it seems trivial (to call `_encode_comparison`), using recursion avoids bugs
            else:
                return _encode_linear(ivarmap, lhs.args, expr.name, rhs, encoding)
        elif lhs.name == "wsum":
            return _encode_linear(
                ivarmap,
                lhs.args[1],
                expr.name,
                rhs,
                encoding,
                weights=lhs.args[0],
            )
        else:
            raise NotImplementedError(f"int2bool: comparison with lhs {lhs} not (yet?) supported")

    else:
        raise NotImplementedError(f"int2bool: non-comparison {expr} not (yet?) supported")


def _encode_int_var(ivarmap, x, encoding):
    """Return encoding of integer variable `x` and its domain constraints (if newly encoded)."""
    if isinstance(x, (BoolVal, _BoolVarImpl)):
        raise TypeError
    elif x.name in ivarmap:  # already encoded
        return ivarmap[x.name], []
    else:
        if encoding == "direct":
            ivarmap[x.name] = IntVarEncDirect(x)
        elif encoding == "order":
            ivarmap[x.name] = IntVarEncOrder(x)
        elif encoding == "binary":
            ivarmap[x.name] = IntVarEncLog(x)
        else:
            raise NotImplementedError(encoding)

        return (ivarmap[x.name], ivarmap[x.name].encode_domain_constraint())


def _encode_linear(ivarmap, xs, cmp, rhs, encoding, weights=None, check_bounds=True):
    """
    Convert a linear constraint to a pseudo-boolean constraint.

    Returns (newexpr, newcons)
    """
    if weights is None:
        weights = len(xs) * [1]

    # Check for trivial sat/unsat, since pysat does not handle those, and we avoid encoding them
    if check_bounds is True:
        # if the bounds are trivially sat/unsat, the value will be set to True/False
        value = None
        lb = sum(w * x.lb if w >= 0 else w * x.ub for w, x in zip(weights, xs))
        ub = sum(w * x.ub if w >= 0 else w * x.lb for w, x in zip(weights, xs))

        if cmp in ("==", "!="):
            if lb == rhs == ub:
                value = BoolVal(True)
            elif rhs not in range(lb, ub + 1):
                value = BoolVal(False)
            if cmp == "!=" and value is not None:
                value = ~value
        elif cmp == "<=":
            if ub <= rhs:
                value = BoolVal(True)
            elif lb > rhs:
                value = BoolVal(False)
        elif cmp == ">=":
            if lb >= rhs:
                value = BoolVal(True)
            elif ub < rhs:
                value = BoolVal(False)
        else:
            raise NotImplementedError

        # if trivial sat/unsat, return early
        if value is not None:
            return [value], []

    terms = []
    domain_constraints = []
    for w, x in zip(weights, xs):
        # the linear may contain Boolean as well as integer variables
        if isinstance(x, _BoolVarImpl):
            terms += [(w, x)]
        else:
            x_enc, x_cons = _encode_int_var(ivarmap, x, _decide_encoding(x, cmp, encoding))
            domain_constraints += x_cons
            # Encode the value of the integer variable as PB expression `(b_1*c_1) + ... + k`
            new_terms, k = x_enc.encode_term(w)
            terms += new_terms  # add new terms
            rhs -= k  # subtract constant from both sides

    if len(terms) == 0:
        # the unzip trick does not allow default for 0 length iterables
        lhs = 0
    else:
        # term tuples to two separate lists
        pb_weights, pb_literals = zip(*terms)

        # Revert back to `sum` if we happen to have constructed one
        if all(w == 1 for w in pb_weights):
            lhs = Operator("sum", pb_literals)
        else:
            lhs = Operator("wsum", (pb_weights, pb_literals))

    return [Comparison(cmp, lhs, rhs)], domain_constraints


def _encode_comparison(ivarmap, lhs, cmp, rhs, encoding):
    """Encode integer comparison to PB."""
    # TODO encode_expr should only use encode linear and check for "comparison" there
    encoding = _decide_encoding(lhs, cmp, encoding)
    lhs_enc, domain_constraints = _encode_int_var(ivarmap, lhs, encoding)
    constraints = lhs_enc.encode_comparison(cmp, rhs)
    return constraints, domain_constraints


def _decide_encoding(x, cmp=None, encoding="auto"):
    """Decide encoding of `x` via a simple heuristic based on linear comparator and domain size."""
    if encoding != "auto":
        return encoding
    elif _dom_size(x) >= 100:
        # This heuristic is chosen to be small to favour the binary encoding. This is because the PB encoding (e.g. generalized totalizer, ...) of a direct/order encoded PB constraints is quite inefficient unless the AMO/IC side-constraint is taken into account (which is not the case for pysat/pblib/pysdd).
        return "binary"
    elif cmp in (None, "==", "!="):
        return "direct"  # equalities suit the direct encoding
    else:  # inequalities suit the order encoding
        return "order"


class IntVarEnc(ABC):
    """Abstract base class for integer variable encodings."""

    def __init__(self, x, n, name):
        """Create encoding of integer variable `x` with `n` Boolean variables named by `name`."""
        if _dom_size(x) == 0:
            raise EMPTY_DOMAIN_ERROR

        self._x = x  # the encoded integer variable

        if n == 0:
            # `shape=(0,)` raises exception
            self._xs = cp.cpm_array([])
        else:
            # `x`'s encoding variables
            self._xs = boolvar(shape=(n,), name=name)

    def vars(self):
        """Return the Boolean variables in the encoding."""
        return self._xs

    def _offset(self, d):
        """ "Map domain value `d` to the encoding range (e.g. for the direct encoding to the encoding variable index x==i)"""
        return d - self._x.lb

    def decode(self):
        """Return the integer value of the encoding."""
        terms, k = self.encode_term()
        for weight, lit in terms:
            if lit is None:
                return None
            elif lit.value() is True:
                k += weight
        return k

    @abstractmethod
    def encode_domain_constraint(self):
        """
        Return domain constraints for the encoding.

        Returns:
            List[Expression]: a list of constraints
        """
        pass

    @abstractmethod
    def encode_comparison(self, op, rhs):
        """
        Encode a comparison over the variable: self <op> rhs.

        Args:
            op: The comparison operator ("==", "!=", "<", "<=", ">", ">=")
            rhs: The right-hand side value to compare against

        Returns:
            List[Expression]: a list of constraints
        """
        pass

    @abstractmethod
    def encode_term(self, w=1):
        """Encode `w*self` as a weighted sum of Boolean variables plus a constant.

        Args:
            w: The weight to multiply the variable by

        Returns:
            tuple: (terms, k)  # with `terms` a list of `(weight, literal)` tuples and `k` an integer constant
        """
        pass


class IntVarEncDirect(IntVarEnc):
    """
    Direct (or sparse or one-hot) encoding of an integer variable.

    Uses a Boolean 'equality' variable for each value in the domain.
    """

    def __init__(self, x):
        """Create direct encoding of integer variable `x`."""
        # Requires |dom(x)| Boolean equality variables
        super().__init__(x, _dom_size(x), f"EncDir({x.name})")

    def encode_domain_constraint(self):
        """
        Return consistency constraints.

        Variable x has exactly one value from domain,
        so only one of the Boolean variables can be True
        """
        return [cp.sum(self._xs) == 1]

    def eq(self, d):
        """Return a literal whether x==d."""
        if self._x.lb <= d <= self._x.ub:
            return self._xs[self._offset(d)]
        else:  # don't use `try .. except IndexError` since negative values wrap!
            return BoolVal(False)

    def encode_comparison(self, op, d):
        if op == "==":
            # one yes, hence also rest no, if rhs is not in domain will set all to no
            return [self.eq(d)]
        elif op == "!=":
            return [~self.eq(d)]
        # return _not_and([self.eq(d)])
        elif op == "<=":
            # all higher values are False
            return list(~self._xs[self._offset(d + 1) :])
        elif op == ">=":
            # all lower-or-equal values are False
            return list(~self._xs[: self._offset(d)])
        else:
            raise UNKNOWN_COMPARATOR_ERROR

    def encode_term(self, w=1):
        return [(w * i, b) for i, b in enumerate(self._xs)], self._x.lb * w


class IntVarEncOrder(IntVarEnc):
    """
    Order (or thermometer) encoding of an integer variable.

    Uses a Boolean 'inequality' variable for each value in the domain.
    """

    def __init__(self, x):
        """Create order encoding of integer variable `x`."""
        super().__init__(x, _dom_size(x) - 1, f"EncOrd({x.name})")

    def encode_domain_constraint(self):
        """Return order encoding domain constraint (i.e. encoding variables are sorted in descending order e.g. `111000`)."""
        if len(self._xs) <= 1:
            return []
        # Encode implication chain `x>=d -> x>=d-1` (using `zip` to create a sliding window)
        return [curr.implies(prev) for prev, curr in zip(self._xs, self._xs[1:])]

    def _offset(self, d):
        return d - self._x.lb - 1

    def eq(self, d):
        """Return a conjunction whether x==d."""
        if self._x.lb <= d <= self._x.ub:
            return [self.geq(d), ~self.geq(d + 1)]
            # return cp.all([self.geq(d), ~self.geq(d + 1)])
        else:
            return [BoolVal(False)]

    def geq(self, d):
        """Return a literal whether x>=d."""
        if d <= self._x.lb:
            return BoolVal(True)
        elif d > self._x.ub:
            return BoolVal(False)
        else:
            return self._xs[self._offset(d)]

    def encode_comparison(self, cmp, d):
        if cmp == "==":  # x>=d and x<d+1
            return self.eq(d)
        elif cmp == "!=":  # x<d or x>=d+1
            return [cp.any(~lit for lit in self.eq(d))]
            # TODO we could just use ~cp.all and ~cp.any, but we need to make pysat more robust
            # return [~self.eq(d)]
        elif cmp == ">=":
            return [self.geq(d)]
        elif cmp == "<=":
            return [~self.geq(d + 1)]
        else:
            raise UNKNOWN_COMPARATOR_ERROR

    def encode_term(self, w=1):
        return [(w, b) for b in self._xs], w * self._x.lb


class IntVarEncLog(IntVarEnc):
    """
    Log (or "binary") encoding of an integer variable.

    Uses a Boolean 'bit' variable to represent `x` using the unsigned binary representation offset by its lower bound (e.g. for `x in 5..8`, the assignment `00` maps to `x=5`, and `11` to `x=8`). In other words, it is `k`-offset binary encoding where `k=x.lb`.
    """

    def __init__(self, x):
        """Create binary encoding of integer variable `x`."""
        bits = math.ceil(math.log2(_dom_size(x)))
        super().__init__(x, bits, f"EncBin({x.name})")

    def encode_domain_constraint(self):
        """Return binary encoding domain constraint (i.e. upper bound is respected with `self._x<=self._x.ub`. The lower bound is automatically enforced by offset binary which maps `000.. = self._x.lb`)."""
        # encode directly to avoid bounds check for this seemingly tautological constraint
        return self.encode_comparison("<=", self._x.ub, check_bounds=False)

    def _to_little_endian_offset_binary(self, d):
        """Return offset binary representation of `d` as Booleans in order of increasing significance ("little-endian").

        For more details on offset binary, see the docstring of this class. Note that if e.g. the offset (equal to `self.x.lb`) is 0, then for `d=4` we return `001`. If the offset (i.e. `self.x.lb`) is 1, then it returns `11`, representing 3 in binary, as binary value of 3 + offset of 1 = 4. Note that in this second case, one less bit is returned as we require only 2."""
        # more efficient implementation probably not necessary
        i = self._offset(d)
        if i == 0:
            # otherwise the bitstring will formatted to '0' (which is inconsistent with empty encodings of constants)
            return []
        else:
            # generate bitstring from least to most significant bit
            bitstring = reversed("{0:b}".format(i))
            return (True if bit == "1" else False for bit in bitstring)

    def eq(self, d):
        """Returns a list of literals which in conjunction enforced x==d."""
        if self._x.lb == d == self._x.ub:
            return [BoolVal(True)]
        elif self._x.lb <= d <= self._x.ub:
            # x_i = bit_i for every bit in the representation
            return [
                x if bit else (~x)
                for bit, x in itertools.zip_longest(self._to_little_endian_offset_binary(d), self._xs)
            ]
        else:  # don't use try IndexError since negative values wrap
            return [BoolVal(False)]

    def encode_comparison(self, cmp, d, check_bounds=True):
        if cmp == "==":  # x>=d and x<d+1
            return self.eq(d)
        elif cmp == "!=":  # x<d or x>=d+1
            return [cp.any(~lit for lit in self.eq(d))]
            # return [cp.any(~x for x in self.eq(d))]
        elif cmp in (">=", "<="):
            # TODO lexicographic encoding might be more effective, but currently we just use the PB encoding
            constraint, domain_constraints = _encode_linear(
                {self._x.name: self}, [self._x], cmp, d, None, check_bounds=check_bounds
            )
            assert domain_constraints == [], (
                f"{self._x} should have already been encoded, so no domain constraints should be returned"
            )
            return constraint
        else:
            raise UNKNOWN_COMPARATOR_ERROR

    def encode_term(self, w=1):
        return [(w * (2**i), b) for i, b in enumerate(self._xs)], w * self._x.lb


def _dom_size(x):
    return x.ub + 1 - x.lb
